/*
* JBoss, Home of Professional Open Source
* Copyright 2005, JBoss Inc., and individual contributors as indicated
* by the @authors tag. See the copyright.txt in the distribution for a
* full listing of individual contributors.
*
* This is free software; you can redistribute it and/or modify it
* under the terms of the GNU Lesser General Public License as
* published by the Free Software Foundation; either version 2.1 of
* the License, or (at your option) any later version.
*
* This software is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
* Lesser General Public License for more details.
*
* You should have received a copy of the GNU Lesser General Public
* License along with this software; if not, write to the Free
* Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
* 02110-1301 USA, or see the FSF site: http://www.fsf.org.
*/
package org.jboss.profiler.aop.monitoring.model;

import java.io.Serializable;

/**
 * Represents a Value on a given period of time
 *
 * @author Clebert Suconic
 * @created August 6, 2004
 */
public class TimeValue implements Serializable, Comparable {

    public TimeValue(long timeKey) {
        this.timeKey = timeKey;
    }

    /**
     * The key for this TimeInterval. Actually i'm using
     * System.currentTimeMillis() dividing by the desired interval to find the
     * actual timestamp
     */
    long timeKey;

    /** The start of the time, without the dividing */
    long timeStart;

    /**
     * How many callins have being made here
     */
    long callings;

    /**
     * How many callins have being made here
     */
    long maxTime;

    /**
     * The minimal ammount of time used by the application
     */
    long minTime = -1;

    /**
     * The current number of stacked callings on this method
     */
    long currentStack;

    /**
     * The maximum number of stacks reached by this interval
     */
    long maxStack;

    /**
     * Total time spent on this value
     */
    long totalTime;

    /**
     * Total cpu time spent on this value
     */
    long totalClock;

    /**
     * @return Returns the currentStack.
     */
    public long getCurrentStack() {
        return currentStack;
    }

    public long getAvgTime() {
        if (callings==0) {
            return 0;
        }
        else {
            return totalTime/callings;
        }
    }

    /**
     * @return Returns the callings.
     */
    public long getCallings() {
        return callings;
    }

    /**
     * @return Returns the maxStack.
     */
    public long getMaxStack() {
        return maxStack;
    }

    /**
     * @return Returns the maxTime.
     */
    public long getMaxTime() {
        return maxTime;
    }

    /**
     * @return Returns the minTime.
     */
    public long getMinTime() {
        return minTime;
    }

    /**
     * @return Returns the timeKey.
     */
    public long getTimeKey() {
        return timeKey;
    }

    /**
     * @return Returns the totalClock.
     */
    public long getTotalClock() {
        return totalClock;
    }

    /**
     * @return Returns the time.
     */
    public long getTotalTime() {
        return totalTime;
    }

    /**
     * Equals based on timeKey
     *
     */
    public boolean equals(Object obj) {
        return timeKey == ((TimeValue)obj).timeKey;
    }

    /**
     * Even we are using timeKey as hashCode and timeKey is a mutable method,
     * we are not chaging values for elements inserted on Hashtables.
     * We are enabling that change just to help the search over that hashtable.
     */
    public int hashCode() {
        // same implementation as java.lang.Long
    	return (int)(timeKey ^ (timeKey  >>> 32));
    }


    /**
     * Description of the Method
     */
    public synchronized void openStack() {
        currentStack++;
        if (currentStack > maxStack) {
            maxStack = currentStack;
        }
    }

    /**
     * Close the current Stack measuring
     */
    public synchronized void closeStack() {
        currentStack--;
        if (currentStack<0) {
            currentStack=0;
        }
    }

    /**
     * Close the current measuring and calls addCall
     * */
    public synchronized void closeStack(long time, long cpuTime) {
        closeStack();
        addCall(time,cpuTime);
    }


    /**
     * @param cpuTime
     *            The CPUClock needs to be captured from OS/JDK through JNI.
     *            It's a optional feature.
     * @param time
     *            The feature to be added to the Call attribute
     */
    public synchronized void addCall(long time, long cpuTime) {
        callings++;

        totalClock += cpuTime;
        totalTime += time;

        if (time > maxTime) {
            maxTime = time;
        }

        if (minTime != -1 || time < minTime) {
            minTime = time;
        }
    }

    /**
     * @param timeKey The timeKey to set.
     */
    public void setTimeKey(long timeKey) {
        this.timeKey = timeKey;
    }

    /**
     * @return Returns the timeStart.
     */
    public long getTimeStart() {
        return timeStart;
    }
    /**
     * @param timeStart The timeStart to set.
     */
    public void setTimeStart(long timeStart) {
        this.timeStart = timeStart;
    }

    /* (non-Javadoc)
     * @see java.lang.Comparable#compareTo(java.lang.Object)
     */
    public int compareTo(Object o) {
        if (timeKey<((TimeValue)o).timeKey) {
            return -1;
        } else
        if (timeKey>((TimeValue)o).timeKey) {
            return 1;
        } else {
            return 0;
        }
    }
}
