/*
* JBoss, Home of Professional Open Source
* Copyright 2005, JBoss Inc., and individual contributors as indicated
* by the @authors tag. See the copyright.txt in the distribution for a
* full listing of individual contributors.
*
* This is free software; you can redistribute it and/or modify it
* under the terms of the GNU Lesser General Public License as
* published by the Free Software Foundation; either version 2.1 of
* the License, or (at your option) any later version.
*
* This software is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
* Lesser General Public License for more details.
*
* You should have received a copy of the GNU Lesser General Public
* License along with this software; if not, write to the Free
* Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
* 02110-1301 USA, or see the FSF site: http://www.fsf.org.
*/

package org.jboss.profiler.engine;

import org.jboss.profiler.fileProcessor.*;
import org.jboss.profiler.model.JBPClass;
import org.jboss.profiler.model.JBPMethod;
import org.jboss.profiler.model.JBPProcess;
import org.jboss.profiler.tracerModel.JBTMethod;
import org.jboss.profiler.tracerModel.JBTTransaction;
import org.jboss.profiler.util.SPYConsts;

import java.io.IOException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Vector;

/**
 * @author Yoko Seki (y-seki@sdl-hitachi.co.jp)
 * @author Clebert Suconic (clebert.suconic.jboss.org) (Added locking measurement)
 * */
public class TracerEngine implements SPYConsts
{

    private static int iteractionDetail = 0;

    private RunnerEvent event = new RunnerEvent(this);

    private int counterLine = 0;
    private Vector runnerEventListeners = new Vector();

    private java.io.File outputDirectory;

    private JBPMethod keyMethod = new JBPMethod();
    private JBPClass keyClass = new JBPClass();

    public static int targetMethodID;
    private HashMap currentTransactions;
    private HashMap completeTransactions;
    private HashMap incompleteTransactions;
    private int transactionID;

    private SpyReader reader;

    public TracerEngine(SpyReader reader)
    {
        this.reader = reader;
    }

    public void doTrace(String pid, JBPProcess process)
            throws java.io.IOException
    {
        doTrace(pid, process, 10000);
    }

    public void doTrace(String pid, JBPProcess process, int tickInterval)
            throws java.io.IOException
    {

        System.out.println("Tracer engine has started processing ...");

        transactionID = 0;
        counterLine = 0;
        SpyRegister register = null;

        currentTransactions = new HashMap();
        completeTransactions = new HashMap();
        incompleteTransactions = new HashMap();

        while ((register = reader.getBestRegister()) != null)
        {
            counterLine++;

            if (tickInterval > 0 && counterLine % tickInterval == 0)
            {
                event.setCurrentRegister(counterLine);
                fireProcessTick(event);
            }

            if (register instanceof SpyRegisterEnterMethod)
            {
                process((SpyRegisterEnterMethod) register);
            } else if (register instanceof SpyRegisterExitMethod)
            {
                process((SpyRegisterExitMethod) register);
            } else if (register instanceof SpyRegisterContended)
            {
                process((SpyRegisterContended)register);
            }

            reader.readNextRegister();
        }
        postProcess();
        currentTransactions.clear();

        System.out.println("Tracer engine has stopped processing.");
    }

    protected void process(SpyRegisterContended register) throws IOException {

        JBTTransaction currentTransaction
                = (JBTTransaction) currentTransactions.get(new Integer(register.getThreadId()));

        if (currentTransaction==null)
        {
            // ignore as it hasn't come the time yet where a method was filtered
            return;
        }

        // need to get the latest method on the stack
        JBTMethod method = (JBTMethod)currentTransaction.getMethods().get(currentTransaction.getMethods().size()-1);

        if (register.getContendedType()==SpyRegisterContended.CONTENDED_ENTER)
        {
            method.setLockEnter(register.getTime_t());
            method.setNumberOfLocks(method.getNumberOfLocks()+1);
            currentTransaction.setNumberOfLocks(currentTransaction.getNumberOfLocks()+1);
        }
        else
        {
            System.out.println("A lock was processed");
            method.setLockEntered(register.getTime_t());
            if (method.getLockEnter()!=Long.MAX_VALUE)
            {
                method.setLockTime(method.getLockTime() + (method.getLockEntered()-method.getLockEnter()));
                currentTransaction.setLockTime(currentTransaction.getLockTime() + (method.getLockEntered()-method.getLockEnter()));
            }
            method.setLockEnter(Long.MAX_VALUE);
            method.setLockEntered(Long.MAX_VALUE);
        }
    }

    protected void process(SpyRegisterEnterMethod register)
            throws IOException
    {

        JBTTransaction currentTransaction
                = (JBTTransaction) currentTransactions.get(new Integer(register.getThreadId()));

        if (register.getMethodId() == targetMethodID && currentTransaction==null)
        {
            /*if (currentTransaction != null)
            {
                incompleteTransactions.put(currentTransaction, currentTransaction);
                currentTransactions.remove(new Integer(register.getThreadId()));
            } */

            JBTMethod newMethod = new JBTMethod();
            newMethod.setDataStart(register);
            newMethod.setDepth(1);

            ArrayList methodsList = new ArrayList();
            methodsList.add(newMethod);

            currentTransaction = new JBTTransaction();
            transactionID += 1;
            currentTransaction.setTransactionID(transactionID);
            currentTransaction.setMethods(methodsList);
            currentTransaction.setCurrentMethod(methodsList.size() - 1);
            currentTransaction.setCurrentDepth(1);
            currentTransactions.put(new Integer(register.getThreadId()), currentTransaction);
        }

        else
        {
            if (currentTransaction != null)
            {
                JBTMethod newMethod = new JBTMethod();
                newMethod.setDataStart(register);
                newMethod.setDepth(currentTransaction.getCurrentDepth() + 1);

                ArrayList methodsList = currentTransaction.getMethods();
                methodsList.add(newMethod);
                currentTransaction.setTransactionID(currentTransaction.getTransactionID());
                currentTransaction.setMethods(methodsList);
                currentTransaction.setCurrentMethod(methodsList.size() - 1);
                currentTransaction.setCurrentDepth(newMethod.getDepth());
                currentTransactions.put(new Integer(register.getThreadId()), currentTransaction);
            }
        }
    }

    protected void process(SpyRegisterExitMethod register) throws IOException
    {

        JBTTransaction currentTransaction
                = (JBTTransaction) currentTransactions.get(new Integer(register.getThreadId()));

        if (currentTransaction != null)
        {
            ArrayList methodsList = currentTransaction.getMethods();
            JBTMethod currentMethod = (JBTMethod) methodsList.get(currentTransaction.getCurrentMethod());

            if (register.getMethodId() == currentMethod.getMethodID())
            {
                currentMethod.setDataEnd(register);

                methodsList.set(currentTransaction.getCurrentMethod(), currentMethod);
                currentTransaction.setCurrentDepth(currentTransaction.getCurrentDepth() - 1);

                if (register.getMethodId() == targetMethodID && currentMethod.getDepth()==1)
                {
                    /*JBTTransaction completeTransaction = new JBTTransaction();
                    completeTransaction.setTransactionID(currentTransaction.getTransactionID());
                    completeTransaction.setMethods(methodsList); */
                    completeTransactions.put(currentTransaction, currentTransaction);

                    currentTransactions.remove(new Integer(register.getThreadId()));
                } else
                {
                    currentTransaction.setTransactionID(currentTransaction.getTransactionID());
                    currentTransaction.setMethods(methodsList); //?????

                    loop:
                    for (int i = currentTransaction.getCurrentMethod(); i >= 0; i--)
                    {
                        if (((JBTMethod) currentTransaction.getMethods().get(i)).getDepth()
                                == (currentMethod.getDepth() - 1))
                        {
                            currentTransaction.setCurrentMethod(i);
                            break loop;
                        }
                    }

                    currentTransactions.put(new Integer(register.getThreadId()), currentTransaction);
                }
            } else
            {
                incompleteTransactions.put(currentTransaction, currentTransaction);
                currentTransactions.remove(new Integer(register.getThreadId()));
            }
        }
    }

    protected void postProcess()
    {
        Iterator iter = currentTransactions.values().iterator();
        while (iter.hasNext())
        {
            JBTTransaction currentTransaction = (JBTTransaction) iter.next();
            incompleteTransactions.put(currentTransaction, currentTransaction);
        }
    }

    public HashMap getCompleteTransactions()
    {
        return completeTransactions;
    }

    public HashMap getIncompleteTransactions()
    {
        return incompleteTransactions;
    }

    public void removeRunnerEventListener(RunnerEventListener l)
    {
        synchronized (runnerEventListeners)
        {
            if (runnerEventListeners.contains(l))
            {
                runnerEventListeners.removeElement(l);
            }
        }
    }

    public void addRunnerEventListener(RunnerEventListener l)
    {
        synchronized (runnerEventListeners)
        {
            if (!runnerEventListeners.contains(l))
            {
                runnerEventListeners.addElement(l);
            }
        }
    }

    protected void fireProcessTick(RunnerEvent e)
    {
        Vector listeners = (Vector) runnerEventListeners.clone();
        int count = listeners.size();
        for (int i = 0; i < count; i++)
        {
            ((RunnerEventListener) listeners.elementAt(i)).processTick(e);
        }
    }

    protected void fireProcessWarning(String e)
    {
        Vector listeners = (Vector) runnerEventListeners.clone();
        int count = listeners.size();
        for (int i = 0; i < count; i++)
        {
            ((RunnerEventListener) listeners.elementAt(i)).processWarning(
                    counterLine,
                    e);
        }
    }
}
