/*
* JBoss, Home of Professional Open Source
* Copyright 2005, JBoss Inc., and individual contributors as indicated
* by the @authors tag. See the copyright.txt in the distribution for a
* full listing of individual contributors.
*
* This is free software; you can redistribute it and/or modify it
* under the terms of the GNU Lesser General Public License as
* published by the Free Software Foundation; either version 2.1 of
* the License, or (at your option) any later version.
*
* This software is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
* Lesser General Public License for more details.
*
* You should have received a copy of the GNU Lesser General Public
* License along with this software; if not, write to the Free
* Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
* 02110-1301 USA, or see the FSF site: http://www.fsf.org.
*/
package org.jboss.profiler.fileProcessor;

import java.io.BufferedInputStream;
import java.io.DataInput;
import java.io.DataInputStream;
import java.io.InputStream;


/**
* This class process a jboss-profiler log file.
* It handles the binary format directly.
@author Clebert Suconic
*/
public class SpyFileReader {
    public SpyFileReader(InputStream input, boolean windowsAlignment){
        this.input= input;
        this.windowsAlignment=windowsAlignment;
        if (windowsAlignment) {
            dataInput = new WindowsDataInputStream(new BufferedInputStream(input,10*1024));
        } else {
            dataInput = new DataInputStream(new BufferedInputStream(input,10*1024));
        }
    }

    // These can be considered cache of objects...
    // Doing a profiler of the profiler, I've found that
    // those objects were consuming a lot of time to process.
    private SpyRegisterBeginThread beginThread = new SpyRegisterBeginThread();
    private SpyRegisterEndThread endThread = new SpyRegisterEndThread();
    private SpyRegisterLoadClass loadClass = new SpyRegisterLoadClass();
    private SpyRegisterLoadMethod loadMethod = new SpyRegisterLoadMethod();
    private SpyRegisterLoadObject loadObject = new SpyRegisterLoadObject();
    private SpyRegisterFreeObject freeObject = new SpyRegisterFreeObject();
    private SpyRegisterGCStart gcstart = new SpyRegisterGCStart();
    private SpyRegisterGCEnd gcend = new SpyRegisterGCEnd();
    private SpyRegisterEnterMethod entermethod = new SpyRegisterEnterMethod();
    private SpyRegisterExitMethod exitmethod = new SpyRegisterExitMethod();
    private SpyRegisterMonitorIn monitorin = new SpyRegisterMonitorIn();
    private SpyRegisterMonitorOut monitorout = new SpyRegisterMonitorOut();
    private SpyRegisterContended registercontended = new SpyRegisterContended();
    private SpyRegisterRawContended registerraw = new SpyRegisterRawContended();

    /** If something wrong happened to the file, the file will be invalidated, and will
     *  not be read any more, (always returning false on readRegister) */
    private boolean invalid=false;



    public boolean isWindowsAlignment(){ return windowsAlignment; }

    public void setWindowsAlignment(boolean windowsAlignment){ this.windowsAlignment = windowsAlignment; }

    public InputStream getInput(){ return input; }

    /**
     * Read the  register
     * @return false if eof
     * */
    public boolean readSpyRegister() {
      try {
        if (invalid) return false;
        if (isEof()) return false;
        if (!spyRegister.read(dataInput)) {
            this.setEof(true);
            return false;
        }

        if (spyRegister.getId()==SpyRegister.CONST_THREADSTART) {
            currentRegister = beginThread;
            //currentRegister = new SpyRegisterBeginThread();
        } else if (spyRegister.getId()==SpyRegister.CONST_THREADEND) {
            currentRegister = endThread;
            //currentRegister = new SpyRegisterEndThread();
        } else if (spyRegister.getId()==SpyRegister.CONST_LOADCLASS) {
            currentRegister = loadClass;
            //currentRegister = new SpyRegisterLoadClass();
        } else if (spyRegister.getId()==SpyRegister.CONST_LOADMETHOD) {
            currentRegister = loadMethod;
            //currentRegister = new SpyRegisterLoadMethod();
        } else if (spyRegister.getId()==SpyRegister.CONST_LOADOBJECT) {
            currentRegister = loadObject;
            //currentRegister = new SpyRegisterLoadObject();
        } else if (spyRegister.getId()==SpyRegister.CONST_UNLOADOBJECT) {
            currentRegister = freeObject;
            //currentRegister = new SpyRegisterFreeObject();
        } else if (spyRegister.getId()==SpyRegister.CONST_GCSTART) {
            currentRegister = gcstart;
            //currentRegister = new SpyRegisterGCStart();
        } else if (spyRegister.getId()==SpyRegister.CONST_GCFINISH) {
            currentRegister = gcend;
            //currentRegister = new SpyRegisterGCEnd();
        } else if (spyRegister.getId()==SpyRegister.CONST_ENTERMETHOD) {
            currentRegister = entermethod;
            //currentRegister = new SpyRegisterEnterMethod();
        } else if (spyRegister.getId()==SpyRegister.CONST_EXITMETHOD) {
            currentRegister = exitmethod;
            //currentRegister = new SpyRegisterExitMethod();
        } else if (spyRegister.getId()==SpyRegister.CONST_MONITOR_IN) {
            currentRegister = monitorin;
            //currentRegister = new SpyRegisterMonitorIn();
        } else if (spyRegister.getId()==SpyRegister.CONST_MONITOR_OUT) {
            currentRegister = monitorout;
            //currentRegister = new SpyRegisterMonitorOut();
        } else if (spyRegister.getId()==SpyRegister.CONST_CONTENDED_ENTER) {
            currentRegister = registercontended;
            //currentRegister = new SpyRegisterContended();
            ((SpyRegisterContended)currentRegister).setContendedType(SpyRegisterContended.CONTENDED_ENTER);
        } else if (spyRegister.getId()==SpyRegister.CONST_CONTENDED_ENTERED) {
            currentRegister = registercontended;
            //currentRegister = new SpyRegisterContended();
            ((SpyRegisterContended)currentRegister).setContendedType(SpyRegisterContended.CONTENDED_ENTERED);
        } else if (spyRegister.getId()==SpyRegister.CONST_CONTENDED_EXIT) {
            currentRegister = registercontended;
            //currentRegister = new SpyRegisterContended();
            ((SpyRegisterContended)currentRegister).setContendedType(SpyRegisterContended.CONTENDED_EXIT);
        } else if (spyRegister.getId()==SpyRegister.CONST_RAW_CONTENDED_ENTER) {
            currentRegister = registerraw;
            //currentRegister = new SpyRegisterRawContended();
            ((SpyRegisterRawContended)currentRegister).setContendedType(SpyRegisterRawContended.CONTENDED_ENTER);
        } else if (spyRegister.getId()==SpyRegister.CONST_RAW_CONTENDED_ENTERED) {
            currentRegister = registerraw;
            //currentRegister = new SpyRegisterRawContended();
            ((SpyRegisterRawContended)currentRegister).setContendedType(SpyRegisterRawContended.CONTENDED_ENTERED);
        } else if (spyRegister.getId()==SpyRegister.CONST_RAW_CONTENDED_EXIT) {
            currentRegister = registerraw;
            //currentRegister = new SpyRegisterRawContended();
            ((SpyRegisterRawContended)currentRegister).setContendedType(SpyRegisterRawContended.CONTENDED_EXIT);
        } else {
            String strMessage = "Invalid record = "+ spyRegister.getId();
            if (currentRegister!=null) strMessage+=" last record type = " + currentRegister.getId();
            else strMessage += " First record readed";
            throw new java.io.IOException (strMessage);
        }

        currentRegister.read(dataInput,spyRegister);

        return true;
      }
      catch (Exception e) {
        //System.out.println("Error on SpyFileReader - " + e.toString());
        //Actually this is not an error... gzip CRC function has static data declared, and because of that
        //it doesn't work well with multi-thread applications, so the CRC is invalid sometimes.
        invalid=true;
        setEof(true);
        return false;
      }
    }

    public boolean isEof(){ return eof; }

    public void setEof(boolean eof){ this.eof = eof; }

    public SpyRegister getCurrentRegister(){
            return currentRegister;
        }

    //public void readDesc()

    private boolean windowsAlignment;
    private InputStream input;
    private DataInput dataInput;

    /** SpyRegister used to read common data in registers*/
    private SpyRegister spyRegister = new SpyRegister();

    /** Will receive the instance which is being processed */
    private SpyRegister currentRegister = null;

    private boolean eof = false;
}
