/*
* JBoss, Home of Professional Open Source
* Copyright 2005, JBoss Inc., and individual contributors as indicated
* by the @authors tag. See the copyright.txt in the distribution for a
* full listing of individual contributors.
*
* This is free software; you can redistribute it and/or modify it
* under the terms of the GNU Lesser General Public License as
* published by the Free Software Foundation; either version 2.1 of
* the License, or (at your option) any later version.
*
* This software is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
* Lesser General Public License for more details.
*
* You should have received a copy of the GNU Lesser General Public
* License along with this software; if not, write to the Free
* Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
* 02110-1301 USA, or see the FSF site: http://www.fsf.org.
*/
package org.jboss.profiler.fileProcessor;

import java.io.*;
import java.util.Date;

/**
 * Read a set of SpyFiles (containing the main output and the secondary thread
 * output Files)
 *
 * @author Clebert Suconic
 */
public class SpyReader {
    public boolean isWindowsAlignment() {
        return windowsAlignment;
    }

    public void setWindowsAlignment(boolean windowsAlignment) {
        this.windowsAlignment = windowsAlignment;
    }

    InputStream streamInp = null;

    private boolean windowsAlignment;

    private SpyFileReader[] spyFileReader;

    /** The place where we readed the last register */
    private int currentStream = -1;

    /** CLOCKS_PER_SECOND constant defined into C++ compiler used by the SpyDLL */
    private int clocksPerSecond;

    private Date startDate;

    public SpyReader(InputStream mainInput, InputStream[] threadInput)
            throws IOException {
        this.streamInp = mainInput;
        windowsAlignment = readHeader();
        int numberOfStreams = 1;
        if (threadInput == null) {
            threadInput = new InputStream[0];
        }
        numberOfStreams += threadInput.length;
        spyFileReader = new SpyFileReader[numberOfStreams];
        spyFileReader[0] = new SpyFileReader(mainInput, windowsAlignment);
        for (int i = 0; i < threadInput.length; i++) {
            spyFileReader[i + 1] = new SpyFileReader(threadInput[i],
                    windowsAlignment);
        }

        for (int i = 0; i < spyFileReader.length; i++) {
            spyFileReader[i].readSpyRegister();
        }
    }

    /**
     * Find the best register to be read from all files
     *
     * @return @throws
     *         IOException
     */
    public SpyRegister getBestRegister() throws java.io.IOException {
        long current_clock = 0;
        SpyRegister regTmp = null;
        currentStream = -1;

        for (int i = 0; i < spyFileReader.length; i++) {
            if (spyFileReader[i].isEof())
                continue;

            regTmp = spyFileReader[i].getCurrentRegister();
            if (current_clock == 0) {
                current_clock = regTmp.getElapsedClock();
                currentStream = i;
            } else {
                if (current_clock > regTmp.getElapsedClock()) {
                    current_clock = regTmp.getElapsedClock();
                    currentStream = i;
                }
            }
        }

        if (currentStream >= 0) {
            SpyRegister register = spyFileReader[currentStream]
                    .getCurrentRegister();
            return register;
        } else {
            return null;
        }
    }

    public void readNextRegister() throws IOException {
        if (currentStream >= 0) {
            spyFileReader[currentStream].readSpyRegister();
        }
    }

    private boolean compareIdentity(byte[] progId, byte[] testCondition) {
        for (int i = 0; i < testCondition.length; i++) {
            if (progId[i] != testCondition[i])
                return false;
        }
        return true;
    }

    /**
     * @return true if needs to use WindowsDataInputStream, false in another
     *         case
     */
    protected boolean readHeader() throws IOException {
        byte[] progId = new byte[6];
        int bytesRead = streamInp.read(progId);
        if (bytesRead < 0 || bytesRead != progId.length) {
            throw new IOException("Unexpected EOF");
        }

        // An old version before the profiler came to sourceforge used a
        // different progId. This is to keep a compatibility with old generated
        // files. (Will be removed in the future)
        byte oldIdentity[] = { 'S', 'P', 'Y', 'S', 'R', 'V' };
        byte newIdentity[] = { 'D', 'P', 'I', 'N', 'S', 'P' };

        if (!compareIdentity(progId, oldIdentity)
                && !compareIdentity(progId, newIdentity)) {
            throw new IOException("Unexpected file format");
        }

        DataInputStream dataInput = new DataInputStream(streamInp);
        // readNumber is expected to be 255. If is not we need to conver
        // alignment against windows / intel.
        int readNumber = dataInput.readInt();

        long time_t = 0;
        if (readNumber != 255) {
            WindowsDataInputStream tmpInp = new WindowsDataInputStream(
                    streamInp);
            clocksPerSecond = 1;
            time_t = tmpInp.readLong();
        } else {
            clocksPerSecond = 1;
            time_t = dataInput.readLong();
        }

        SpyRegister.startOfProcess = time_t;

        startDate = new Date(time_t);

        return readNumber != 255;
    }

    /** CLOCKS_PER_SECOND constant defined into C++ compiler used by the SpyDLL */
    public int getClocksPerSecond() {
        return clocksPerSecond;
    }

    public Date getStartDate() {
        return startDate;
    }

    public void setStartDate(Date startDate) {
        this.startDate = startDate;
    }
}
