/*
* JBoss, Home of Professional Open Source
* Copyright 2005, JBoss Inc., and individual contributors as indicated
* by the @authors tag. See the copyright.txt in the distribution for a
* full listing of individual contributors.
*
* This is free software; you can redistribute it and/or modify it
* under the terms of the GNU Lesser General Public License as
* published by the Free Software Foundation; either version 2.1 of
* the License, or (at your option) any later version.
*
* This software is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
* Lesser General Public License for more details.
*
* You should have received a copy of the GNU Lesser General Public
* License along with this software; if not, write to the Free
* Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
* 02110-1301 USA, or see the FSF site: http://www.fsf.org.
*/
package org.jboss.profiler.filecollection;

import java.io.BufferedInputStream;
import java.io.ByteArrayInputStream;
import java.io.DataInput;
import java.io.DataInputStream;
import java.io.DataOutput;
import java.io.DataOutputStream;
import java.io.Externalizable;
import java.io.File;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.OutputStream;
import java.io.RandomAccessFile;
import java.nio.channels.Channels;
import java.nio.channels.FileChannel;
import java.util.Collection;
import java.util.Iterator;
import java.util.List;
import java.util.ListIterator;
import java.util.RandomAccess;

class MyByteOutputStream extends OutputStream
{
    byte[] buf;
    int count=0;
    public MyByteOutputStream (byte[] byteArray)
    {
        this.buf = byteArray;
    }

    public byte[] toByteArray()
    {
        return this.buf;
    }

    public void reset()
    {
        count=0;
    }

    public void write(int b) throws IOException
    {
        if (count>=buf.length)
        {
            throw new RuntimeException("Invalid register size");
        }

        buf[count++] = (byte)b;
    }


}


class Index implements FileCollectionInterface
{
    public static final int SIZE = 16;

    long posStart;

    long posEnd;

    public Index()
    {
        super();
    }

    /**
     * @param posStart
     * @param posEnd
     */
    public Index(long posStart, long posEnd)
    {
        super();
        this.posStart = posStart;
        this.posEnd = posEnd;
    }

    /**
     * @return
     */
    public long getPosEnd()
    {
        return posEnd;
    }

    /**
     * @return
     */
    public long getPosStart()
    {
        return posStart;
    }

    public void readObject(DataInput ch) throws IOException
    {
        posStart = ch.readLong();
        posEnd = ch.readLong();
    }

    /**
     * @param posEnd
     */
    public void setPosEnd(long posEnd)
    {
        this.posEnd = posEnd;
    }

    /**
     * @param posStart
     */
    public void setPosStart(long posStart)
    {
        this.posStart = posStart;
    }

    public void writeObject(DataOutput ch) throws IOException
    {
        ch.writeLong(posStart);
        ch.writeLong(posEnd);
    }


}

/**
 * Manipulates registers on files working like a collection. The elements that
 * are being used here, have to implement preferably FileCollectionInterface or Externalizable (do not call supper.writeExternal on that case)

 * @author Clebert Suconic
 */
public class FileCollection implements Collection, List, RandomAccess
{
    Index indCache = new Index();

    CountInputStream countInput;


    private class Itr implements Iterator
    {

        public Itr()
        {
            try
            {
                FileChannel fileChannel = registersChannel.getChannel();
                fileChannel.position(0);

                countInput = new CountInputStream(new BufferedInputStream(Channels.newInputStream(fileChannel)),0);
                inputStream = new DataInputStream(countInput);

            } catch (Exception e)
            {
                throw new RuntimeException(e);
            }

        }

        DataInputStream inputStream;

        /**
         * Index of element to be returned by subsequent call to next.
         */
        int cursor = 0;

        /**
         * Index of element returned by most recent call to next or previous.
         * Reset to -1 if this element is deleted by a call to remove.
         */
        int lastRet = -1;

        public boolean hasNext()
        {
            return cursor != size();
        }

        public Object next()
        {
            if (!hasNext())
                throw new java.util.NoSuchElementException();

            lastRet = cursor++;
            Object o = createInstance();
            try
            {
                long oldPos = countInput.getCurrentCount();
                if (o instanceof FileCollectionInterface)
                {
                    ((FileCollectionInterface) o).readObject(inputStream);
                } else if (o instanceof Externalizable)
                {
                    ((Externalizable) o).readExternal(new ObjectInputStream(
                            inputStream));
                }
                if (fixedSize>0)
                {
                    fixSizeRead(oldPos);
                }
                return o;
            } catch (Exception e)
            {
                e.printStackTrace();
                throw new RuntimeException("IO Exception at FileCollection - "
                        + e.toString());
            }
        }

        private void fixSizeRead(final long oldIndex) throws IOException
        {
            long currentSize =  countInput.getCurrentCount() - oldIndex;

            if (currentSize>fixedSize)
            {
                throw new RuntimeException("Invalid register size");
            }

            // Fixing the size of the register
            for (;currentSize<fixedSize;currentSize++)
            {
                inputStream.read();
            }
        }


        public void remove()
        {
            throw new UnsupportedOperationException();
        }

    }

    /** If using a fixed registers */
    int fixedSize=0;




    byte[] byteBuffer;
    ByteArrayInputStream byteInput;
    MyByteOutputStream byteOut;

    DataInputStream dataInputBuffer;
    DataOutputStream dataOutputBuffer;

    /**
     * A place where the key for acessing the registers will be found
     */
    private RandomAccessFile indexChannel;

    /** The class that is being managed by this FileCollection */
    private Class managedClass;

    /**
     * A place where the registers will be write
     */
    private RandomAccessFile registersChannel;

    private File registersFile;
    private File indexFile;

    /**
     * if we are reusing instances this member will have that cached instances
     * for read operations
     */
    private Object reusedInstance = null;

    /** Reuse instances on read operations making a cache of those objects? */
    private boolean reuseInstances = false;

    /** The number of elements that is being stored */
    private int size = 0;

    public FileCollection(RandomAccessFile registersChannel,
            RandomAccessFile indexChannel, Class managedClass,
            boolean reuseInstances, int fixedSize) throws java.io.IOException
    {
        this.fixedSize=fixedSize;
        this.managedClass = managedClass;
        this.registersChannel = registersChannel;
        this.indexChannel = indexChannel;
        this.calculateSize();
        initializeBuffers();
    }

    public FileCollection(File registersFile,
            File indexFile, Class managedClass,
            boolean reuseInstances,int fixedSize) throws java.io.IOException
    {
        this.registersFile = registersFile;
        this.indexFile = indexFile;
        this.managedClass = managedClass;
        this.registersChannel = new RandomAccessFile(registersFile,"rw");
        this.fixedSize=fixedSize;

        if (fixedSize==0)
        {
            this.indexChannel = new RandomAccessFile(indexFile,"rw");
        }
        calculateSize();
        initializeBuffers();
    }

    private void initializeBuffers()
    {
        if (this.fixedSize!=0)
        {
            this.byteBuffer = new byte[fixedSize];
            this.byteOut = new MyByteOutputStream(byteBuffer);
            this.byteInput = new ByteArrayInputStream(byteBuffer);

            dataInputBuffer = new DataInputStream(byteInput);
            dataOutputBuffer = new DataOutputStream(byteOut);
            DataOutputStream dataOutputBuffer;

        }
    }

    void calculateSize() throws IOException
    {
        if (fixedSize==0)
        {
            this.size = (int) (indexChannel.length() / Index.SIZE);
        }
        else
        {
            this.size = (int)registersChannel.length() / fixedSize;
        }
    }

    /*
     * (non-Javadoc)
     *
     * @see java.util.List#add(int, java.lang.Object)
     */
    public void add(int index, Object element)
    {
        throw new RuntimeException("not implemented!");

    }

    /*
     * (non-Javadoc)
     *
     * @see java.util.Collection#add(java.lang.Object)
     */
    public boolean add(Object o)
    {
        try
        {
            long startPosition = this.registersChannel.getFilePointer();
            registersChannel.seek(registersChannel.length());
            write(o);
            long endPosition = this.registersChannel.getFilePointer();
            addIndex(startPosition, endPosition);

            size++;
            return true;
        } catch (IOException e)
        {
            e.printStackTrace();
            throw new RuntimeException("IO Error on FileCollection - "
                    + e.toString());
        }
    }

    /*
     * (non-Javadoc)
     *
     * @see java.util.Collection#addAll(java.util.Collection)
     */
    public boolean addAll(Collection c)
    {
        throw new RuntimeException("Not implemented yet!");
    }

    /*
     * (non-Javadoc)
     *
     * @see java.util.List#addAll(int, java.util.Collection)
     */
    public boolean addAll(int index, Collection c)
    {
        throw new RuntimeException("not implemented!");
    }

    /*
     * (non-Javadoc)
     *
     * @see java.util.Collection#clear()
     */
    public void clear()
    {
        throw new RuntimeException("Method not implemented");

    }

    public void close()
    {
        if (indexFile==null || registersFile==null )
        {
            throw new RuntimeException("File wasn't opened inside FileCollection");
        }

        try
        {
            registersChannel.close();
            indexChannel.close();
        } catch (IOException e)
        {
            throw new RuntimeException(e);
        }
    }

    /*
     * (non-Javadoc)
     *
     * @see java.util.Collection#contains(java.lang.Object)
     */
    public boolean contains(Object o)
    {
        throw new RuntimeException("Not implemented yet!");
    }

    /*
     * (non-Javadoc)
     *
     * @see java.util.Collection#containsAll(java.util.Collection)
     */
    public boolean containsAll(Collection c)
    {
        throw new RuntimeException("Not implemented yet!");
    }

    /**
     * Creates the instance to be readed, allowing the use of a cached instance.
     *
     * @return
     */
    Object createInstance()
    {
        try
        {
            if (this.reuseInstances)
            {
                if (this.reusedInstance != null)
                    return this.reusedInstance;
                else
                {
                    this.reusedInstance = managedClass.newInstance();
                    return this.reusedInstance;
                }
            } else
            {
                return managedClass.newInstance();
            }
        } catch (Exception e)
        {
            e.printStackTrace();
            throw new RuntimeException(e);
        }
    }

    /*
     * (non-Javadoc)
     *
     * @see java.util.List#get(int)
     */
    public Object get(int index)
    {
        try
        {
            gotoPosition(index);
            return read();
        } catch (IOException e)
        {
            e.printStackTrace();
            throw new RuntimeException(e);
        }

    }

    /*
     * Gets an object but uses it as source to read it
     */
    public Object get(int index, Object obj)
    {
        try
        {
            gotoPosition(index);
            return read(obj);
        } catch (IOException e)
        {
            e.printStackTrace();
            throw new RuntimeException(e);
        }

    }

    /**
     * @return
     */
    public Class getManagedClass()
    {
        return managedClass;
    }

    /**
     * @return
     */
    public RandomAccessFile getRegistersChannel()
    {
        return registersChannel;
    }

    /** If readonly means we don't need the Index return */
    private Index gotoPosition(final int index) throws IOException
    {
        final Index ind = readIndex(index);
        final long posStart = ind.getPosStart();


        this.registersChannel.seek(posStart);
        return ind;
    }

    public long locatePositionIndex (final int index) throws IOException
    {
        if (fixedSize>0)
        {
            return (long)fixedSize * (long)index;
        } else
        {
            final Index ind = readIndex(index);
            return ind.getPosStart();
        }
    }


    /*
     * (non-Javadoc)
     *
     * @see java.util.List#indexOf(java.lang.Object)
     */
    public int indexOf(Object o)
    {
        throw new RuntimeException("not implemented");
    }

    /*
     * (non-Javadoc)
     *
     * @see java.util.Collection#isEmpty()
     */
    public boolean isEmpty()
    {
        return false;
    }

    /*
     * (non-Javadoc)
     *
     * @see java.util.Collection#iterator()
     */
    public Iterator iterator()
    {
        return new Itr();
    }

    /*
     * (non-Javadoc)
     *
     * @see java.util.List#lastIndexOf(java.lang.Object)
     */
    public int lastIndexOf(Object o)
    {
        throw new RuntimeException("not implemented");
    }

    /*
     * (non-Javadoc)
     *
     * @see java.util.List#listIterator()
     */
    public ListIterator listIterator()
    {
        throw new RuntimeException("not implemented");
    }

    /*
     * (non-Javadoc)
     *
     * @see java.util.List#listIterator(int)
     */
    public ListIterator listIterator(int index)
    {
        throw new RuntimeException("not implemented");
    }

    protected Object read()
    {
        return read(null);
    }

    private Object read(Object o)
    {
        if (o==null)
        {
            o = createInstance();
        }
        try
        {
            if (o instanceof FileCollectionInterface)
            {
                if (fixedSize!=0)
                {
                    registersChannel.read(this.byteBuffer);
                    byteInput.reset();
                    ((FileCollectionInterface) o).readObject(dataInputBuffer);
                } else
                {
                    ((FileCollectionInterface) o).readObject(registersChannel);
                }
            } else
            {
                throw new RuntimeException("FileCollection only accepts FileCollectionInterface");
            }
            return o;
        } catch (Exception e)
        {
            e.printStackTrace();
            throw new RuntimeException("IO Exception at FileCollection - "
                    + e.toString());
        }
    }

    /*
     * (non-Javadoc)
     *
     * @see java.util.List#remove(int)
     */
    public Object remove(int index)
    {
        throw new RuntimeException("not implemented");
    }

    /*
     * (non-Javadoc)
     *
     * @see java.util.Collection#remove(java.lang.Object)
     */
    public boolean remove(Object o)
    {
        throw new RuntimeException("Not implemented yet!");
    }

    /*
     * (non-Javadoc)
     *
     * @see java.util.Collection#removeAll(java.util.Collection)
     */
    public boolean removeAll(Collection c)
    {
        throw new RuntimeException("Not implemented yet!");
    }

    /*
     * (non-Javadoc)
     *
     * @see java.util.Collection#retainAll(java.util.Collection)
     */
    public boolean retainAll(Collection c)
    {
        throw new RuntimeException("Not implemented yet!");
    }

    /*
     * (non-Javadoc)
     *
     * @see java.util.List#set(int, java.lang.Object)
     */
    public Object set(int index, Object element)
    {
        try
        {
            Index ind = gotoPosition(index);
            write(element);
            if (this.registersChannel.getFilePointer() > ind.getPosEnd())
            {
                throw new RuntimeException(
                        "The length of a register cannot be changed. pointer="+this.registersChannel.getFilePointer() + " and end = " + ind.getPosEnd());
            }
            return element;
        } catch (Exception e)
        {
            e.printStackTrace();
            throw new RuntimeException(e);
        }
    }

    /**
     * @param managedClass
     */
    public void setManagedClass(Class managedClass)
    {
        this.managedClass = managedClass;
    }

    /**
     * @param registersChannel
     */
    public void setRegistersChannel(RandomAccessFile registersChannel)
    {
        this.registersChannel = registersChannel;
    }

    /*
     * (non-Javadoc)
     *
     * @see java.util.Collection#size()
     */
    public int size()
    {
        return this.size;
    }

    /*
     * (non-Javadoc)
     *
     * @see java.util.List#subList(int, int)
     */
    public List subList(int fromIndex, int toIndex)
    {
        throw new RuntimeException("not implemented");
    }

    /*
     * (non-Javadoc)
     *
     * @see java.util.Collection#toArray()
     */
    public Object[] toArray()
    {
        Object[] objs = new Object[this.size()];

        for (int i = 0; i < objs.length; i++)
        {
            objs[i] = this.get(i);
        }

        return objs;
    }

    /*
     * (non-Javadoc)
     *
     * @see java.util.Collection#toArray(java.lang.Object[])
     */
    public Object[] toArray(Object[] a)
    {
        throw new RuntimeException(
                "This method cannot be used as it would use an huge ammount of memory");
    }

    protected synchronized void write(Object o)
    {
        try
        {
            long oldIndex = registersChannel.getFilePointer();
            if (o instanceof FileCollectionInterface)
            {
                // Implementation for performance
                if (fixedSize>0)
                {
                    byteOut.reset();
                    ((FileCollectionInterface) o).writeObject(dataOutputBuffer);
                    registersChannel.write(byteBuffer);
                } else
                {
                    ((FileCollectionInterface) o).writeObject(registersChannel);
                }
            } else
            {
                throw new RuntimeException("FileCollection only accepts FileCollectionInterface");
            }
        } catch (IOException e)
        {
            e.printStackTrace();
            throw new RuntimeException("IO Exception at FileCollection - "
                    + e.toString());
        }
    }

    private void fixSize(long oldIndex) throws IOException
    {
        long currentSize = registersChannel.getFilePointer() - oldIndex;
        if (currentSize>fixedSize)
        {
            throw new RuntimeException("Invalid register size");
        }

        // Fixing the size of the register
        for (;currentSize<fixedSize;currentSize++)
        {
            registersChannel.write(0);
        }
    }

    private Index readIndex(int position) throws IOException
    {
        Index ind = indCache;

        if (fixedSize==0)
        {
            ind = new Index();
        }

        return readIndex(position, ind);
    }

    private Index readIndex(int position, Index ind) throws IOException
    {
        if (fixedSize==0)
        {
            long posInd = position * Index.SIZE;

            indexChannel.seek(posInd);

            ind.readObject(indexChannel);
        } else
        {
            ind.setPosStart((long)position * fixedSize);
            ind.setPosEnd(ind.getPosStart() + fixedSize);
        }
        return ind;
    }

    void addIndex(long posStart, long posEnd) throws IOException
    {
        if (fixedSize==0)
        {
            Index ind = new Index(posStart, posEnd);
            indexChannel.seek(indexChannel.length());
            ind.writeObject(indexChannel);
        }
    }


    /**
     * @return Returns the fixedSize.
     */
    public int getFixedSize()
    {
        return fixedSize;
    }

    /**
     * @param fixedSize The fixedSize to set.
     */
    public void setFixedSize(int fixedSize)
    {
        this.fixedSize = fixedSize;
    }

    public BulkAdd createBulkAdd() throws Exception
    {
        return new BulkAdd(registersChannel,indexChannel,this);
    }

}

