/*
* JBoss, Home of Professional Open Source
* Copyright 2005, JBoss Inc., and individual contributors as indicated
* by the @authors tag. See the copyright.txt in the distribution for a
* full listing of individual contributors.
*
* This is free software; you can redistribute it and/or modify it
* under the terms of the GNU Lesser General Public License as
* published by the Free Software Foundation; either version 2.1 of
* the License, or (at your option) any later version.
*
* This software is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
* Lesser General Public License for more details.
*
* You should have received a copy of the GNU Lesser General Public
* License along with this software; if not, write to the Free
* Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
* 02110-1301 USA, or see the FSF site: http://www.fsf.org.
*/

package org.jboss.profiler.filecollection;

import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.Iterator;

/**
 * @author csuconic
 */
public class QuickSorter
{
    Object cache1 = null;
    Object cache2 = null;

    int swapint = 0;
    Comparator comparator;

    public QuickSorter(Comparator comparator)
    {
        this.comparator = comparator;
    }

    /**
     * O(1) | O(1)
     */
    private static void isValid(FileCollection a, int startIndex, int finishIndex)
    {
        if (!((startIndex >= 0) && (startIndex <= a.size()) && (finishIndex >= 0) && (finishIndex <= a.size())
                && (startIndex <= finishIndex)))
                {
                    throw new RuntimeException ("invalid quicksort indexes - " + startIndex + "" + finishIndex + " for a collection of " + a.size());
                }
    }

    public boolean isSorted(final FileCollection collection)
    {

        Object old = null;
        Iterator iter = collection.iterator();
        int read=0;
        while (iter.hasNext())
        {
            read++;
            if (read%10000==0)
            {
                System.out.println("isSorted: Read " + read + " registers");
            }
            Object obj = iter.next();
            if (old!=null)
            {
                int result = comparator.compare(old,obj);
                if (comparator.compare(old,obj)>=0)
                {
                    if (result>0)
                    {
                        System.out.println("Isn't ordened because " + old + " < " + obj + " at position " + read);
                        return false;
                    }
                }
            }
            old = obj;
        }

        return true;
    }

    /**
     * O(1) | O(1)
     */
    private static int pivot(final FileCollection collection, final int startIndex, final int endIndex)
    {
        isValid(collection, startIndex,endIndex);
        //assert (e - b) > 1;
        final int m = (startIndex + endIndex) >> 1;
        //assert (m > b) && (m < e);
        return m;
    }

    /**
     * O(1) | O(1)
     */
    private void swap(final FileCollection a, final int i, final int j)
    {
        if (i == j)
            return;

        if ((swapint++%20000)==0)
        {
            System.out.println(swapint + " swaps");
        }

        final Object tmpi = a.get(i,cache1);
        final Object tmpj = a.get(j,cache2);
        a.set(i,tmpj);
        a.set(j,tmpi);
    }

    /**
     * O(1) | O(n)
     */
    private int partition(final FileCollection a, int b, int e, int pivotPosition)
    {
        //assert isValid(a, b, e);
        //assert (e - b) > 1;
        //assert (p >= b) && (p < e);
        final Object v = a.get(pivotPosition);
        --e;
        swap(a, pivotPosition, e);
        for (int d = e; true;)
        {
            --d;
            while (comparator.compare(a.get(b,cache1),v)<0)
                ++b;
            while ((b != d) && (comparator.compare(a.get(d,cache1),v)>0))
                --d;
            if (b >= d)
                break;
            swap(a, b, d);
            ++b;
        }
        swap(a, b, e);
        return b;
    }

    /**
     * O(log n) -> O(n) | O(n log n) -> O(n^2) | not stable
     */
    public void sort(final FileCollection collection, final int startIndex, final int finalIndex)
    {
        //assert isValid(a, b, e);
        if ((finalIndex - startIndex) < 2)
            return;
        final int p = pivot(collection, startIndex, finalIndex);
        final int q = partition(collection, startIndex, finalIndex, p);
        sort(collection, startIndex, q);
        sort(collection, q + 1, finalIndex);
    }

    /**
     * O(log n) -> O(n) | O(n log n) -> O(n^2) | not stable
     */
    public void sort(final FileCollection a)
    {
        this.cache1= a.createInstance();
        this.cache2= a.createInstance();
        if (a.size()< 2)
            return;
        sort(a, 0, a.size());
    }

    public Object binarySearchItem(final FileCollection list, final Object key)
    {
        int position = Collections.binarySearch(list,key,this.comparator);

        if (position<0)
        {
            return null;
        } else
        {
            return list.get(position);
        }

    }

    public ArrayList binarySearchCollection(final FileCollection list, final Object key)
    {
        final ArrayList listReturn = new ArrayList();

        int position = Collections.binarySearch(list,key,this.comparator);

        if (position>0)
        {
            for (int i=position;i>=0;i--)
            {
                Object obj = list.get(i);
                if (comparator.compare(key,obj)==0)
                {
                    listReturn.add(list.get(i));
                }
                else
                {
                    break;
                }
            }

            for (int i=position+1;i<=list.size();i++)
            {
                Object obj = list.get(i);
                if (comparator.compare(key,obj)==0)
                {
                    listReturn.add(list.get(i));
                }
                else
                {
                    break;
                }
            }
        }

        return listReturn;
    }

}
