/*
* JBoss, Home of Professional Open Source
* Copyright 2005, JBoss Inc., and individual contributors as indicated
* by the @authors tag. See the copyright.txt in the distribution for a
* full listing of individual contributors.
*
* This is free software; you can redistribute it and/or modify it
* under the terms of the GNU Lesser General Public License as
* published by the Free Software Foundation; either version 2.1 of
* the License, or (at your option) any later version.
*
* This software is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
* Lesser General Public License for more details.
*
* You should have received a copy of the GNU Lesser General Public
* License along with this software; if not, write to the Free
* Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
* 02110-1301 USA, or see the FSF site: http://www.fsf.org.
*/
package org.jboss.profiler.model;

import java.util.HashMap;
import java.util.Iterator;

/**
 * JBPMethodCount is a class which have at the same time the callings between
 * to methods and the count between those methods. It can be used to count
 * methods callings and objects creations at memory's view
@author Clebert Suconic
*/
public class JBPMethodCount extends JBPGenericCount  {

  private static final long serialVersionUID = 100L;

    /** @associates <{JBPGenericCount}>
     * @label callers*/
    private HashMap countCallers;

    /** @associates <{JBPGenericCount}>
     * @label calleds*/
    private HashMap countCalleds;
    /*# JBPMethod spyMethod; */

    /** @associates <{JBPMemoryCount}>
     * @label created Objects
     * Mostra os objetos criados a partir desta chamada
     * */
    private HashMap countMemory;

    /** This member is used when this class is in role of objects creation */
    protected int released;

    public JBPMethodCount(JBPMethod referencedMethod) {
        super(referencedMethod);
    }


    public JBPMethod getSpyMethod(){ return (JBPMethod)this.getReferencedObject(); }

    public void setSpyMethod(JBPMethod spyMethod){ this.setReferencedObject(spyMethod); }

    public JBPMethodCount getCallerCount(JBPMethod method) {
       JBPMethodCount count = (JBPMethodCount)getCountCallers().get(method);
       if (count==null) {
          count = new JBPMethodCount(method);
          getCountCallers().put(method,count);
          method.addSpyMethodCount(count);
       }
       return count;
    }

    public JBPMethodCount getCalledCount(JBPMethod method) {
       return getCalledCount(method,true);
    }
    public JBPMethodCount getCalledCount(JBPMethod method, boolean create) {
       JBPMethodCount count = (JBPMethodCount)getCountCalleds().get(method);
       if (count==null && create) {
          count = new JBPMethodCount(method);
          getCountCalleds().put(method,count);
          method.addSpyMethodCount(count);
       }
       return count;
    }

  public JBPMemoryCount getCountForClass(JBPClass spyClass) {
     return getCountForClass(spyClass,true);
  }

  /** @return Contador para a classe desejada */
  public JBPMemoryCount getCountForClass(JBPClass spyClass, boolean createIfNotFound) {
      JBPMemoryCount memCount = (JBPMemoryCount)this.getCountMemory().get(spyClass);
      if (memCount==null && createIfNotFound) {
         memCount = new JBPMemoryCount(spyClass);
         this.getCountMemory().put(spyClass,memCount);
      }
      return memCount;
  }

  /** Consolidate subMethods for object creation*/
  public HashMap createConsolidatedObjects() {
     HashMap hashObjects = new HashMap();
     copyCount(hashObjects,this);
     return hashObjects;
  }

  /** Metodo recursivo para consolidacao dos sub-metodos */
  private static void copyCount(HashMap hashObjects, JBPMethodCount methodToCopy) {
     Iterator memoryIterator = methodToCopy.getCountMemory().values().iterator();
     while (memoryIterator.hasNext()) {
        JBPMemoryCount countInstance = (JBPMemoryCount)memoryIterator.next();
        JBPMemoryCount countTarget = (JBPMemoryCount)hashObjects.get(countInstance.getReferencedObject());
        if (countTarget==null) {
           countTarget = new JBPMemoryCount(countInstance.getReferencedObject());
           hashObjects.put(countInstance.getReferencedObject(),countTarget);
        }
        countTarget.sum(countInstance);
     }

     Iterator subMethodIterator = methodToCopy.getCountCalleds().values().iterator();
     while (subMethodIterator.hasNext()) {
         JBPMethodCount subMethod = (JBPMethodCount)subMethodIterator.next();
         copyCount(hashObjects,subMethod);
     }
  }

  /** Array of {@link JBPMemoryCount}*/
  public HashMap getCountMemory() {
    if (countMemory==null) countMemory = new HashMap();
    return countMemory;
  }

  public HashMap getCountCalleds() {
    if (countCalleds==null) countCalleds = new HashMap();
    return countCalleds;
  }
  public HashMap getCountCallers() {
    if (countCallers==null) countCallers = new HashMap();
    return countCallers;
  }

    public void printDebugLevel(int level, String message) {
       for (int i=0;i<level;i++) {
           System.out.print("|");
           for (int j=0;j<3;j++) System.out.print("-");
       }
       System.out.print(message);
       System.out.println(this.getCount() + " chamadas com " + this.getTotalClocks() + " clocks "  + this.getTotalElapsedClock() + " segundos em " + this.getSpyMethod().getFullName()+ ":");
    }

    public void printDebugCalleds(int level) {
       this.printDebugLevel(level,"calleds methods of ");
          Iterator iterClasses = this.getCountMemory().values().iterator();
          while (iterClasses.hasNext()) {
           JBPMemoryCount countMemory = (JBPMemoryCount)iterClasses.next();

           for (int i=0;i<level;i++) {
               System.out.print(">");
               for (int j=0;j<3;j++) System.out.print("*");
           }
           System.out.println(" class = " + ((JBPClass)countMemory.getReferencedObject()).getName() + " com " + countMemory.getCount() + " instancias");
          }

       Iterator iter = getCountCalleds().values().iterator();
       while (iter.hasNext()) {
          JBPMethodCount methodIter = (JBPMethodCount)iter.next();
          methodIter.printDebugCalleds(level+1);

       }
    }

    public void printDebugCallers(int level) {
       this.printDebugLevel(level,"callers methods of");
       Iterator iter = getCountCallers().values().iterator();
       while (iter.hasNext()) {
          JBPMethodCount methodIter = (JBPMethodCount)iter.next();
          methodIter.printDebugCalleds(level+1);
       }
    }


	/**
	 * @return
	 */
	public int getReleased() {
		return released;
	}

	public int getActiveCount() {
		return getCount() - getReleased();
	}

	/**
	 * @param released
	 */
	public void setReleased(int released) {
		this.released = released;
	}

	public void addRelease() {
		released++;
	}

}
