/*
* JBoss, Home of Professional Open Source
* Copyright 2005, JBoss Inc., and individual contributors as indicated
* by the @authors tag. See the copyright.txt in the distribution for a
* full listing of individual contributors.
*
* This is free software; you can redistribute it and/or modify it
* under the terms of the GNU Lesser General Public License as
* published by the Free Software Foundation; either version 2.1 of
* the License, or (at your option) any later version.
*
* This software is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
* Lesser General Public License for more details.
*
* You should have received a copy of the GNU Lesser General Public
* License along with this software; if not, write to the Free
* Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
* 02110-1301 USA, or see the FSF site: http://www.fsf.org.
*/
package org.jboss.profiler.model;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.RandomAccessFile;
import java.io.Serializable;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Date;
import java.util.HashMap;
import java.util.Iterator;

import org.jboss.profiler.filecollection.FileCollection;

/**
 * Contains all the relationships pertinent to the gathered events, like
 * method's graph, profiled events, and garbage collection.  Created after the
 * evaluation of the LogFiles by {@link
 * org.jboss.profiler.engine.LogFileEngine}.  Consider this as the rootObject
 * on this dataModel.
 *
 * @todo The serialization doesn't work on this model, as there is lots of
 * circular references... Needs improvement
 * @author Clebert Suconic
 */
public class JBPProcess
	implements Serializable, org.jboss.profiler.util.SPYConsts {

	private static final long serialVersionUID = 100L;

	/** PK */
	private Date date;

	private java.util.Date finishDate;

	/**
	 *@link aggregation
	 *      @associates <{JBPMemoryState}>
	 */
	private ArrayList memoryStates = new ArrayList();

	private int numberOfGarbageCollects;
	private String pid;
	private JBPThread rootThread = new JBPThread();

	/**
	 *@link aggregation
	 *      @associates <{JBPClass}>
	 */
	private HashMap spyClasses = new HashMap();

	/** @associates <{JBPMethod}>*/
	private HashMap spyMethods = new HashMap();

	/**
	 * Objects created during operations.
	 * This collection contains instances of {@link JBPObject}
	 */
	private transient FileCollection spyObjects = null;

	/**
	 * Objects released during operations.
	 * This collection contains instances of {@link JBPObjectInstance}
	 */
	private transient FileCollection spyObjectsGC = null;

	/**
	 * the key for this HashMap is <@link JBPClass>.
	 * Each element of this HashMap is an instance of {@link HashMap} representing method-roots for that spyClass */
	private HashMap spyObjectTreeByClass = new HashMap();

	/** @associates <{JBPThread}>*/
	private HashMap spyThreads = new HashMap();
	private long totalClocks;
	private long totalElapsedClocks;
	private long totalTimeOfGCMilliseconds;

	public JBPProcess() {
		rootThread.setName("rootThread");
		memoryStates.add(new JBPMemoryState());

                try {
                        this.spyObjects = createFileCollection("objs", JBPObject.class);
                        this.spyObjectsGC = createFileCollection("gcs", JBPObjectInstance.class);
                } catch (IOException e) {
                        throw new RuntimeException ("Cannot setup temporary files");
                }
	}

	/**
	 * @return
	 */
	public Collection getSpyObjects() {
		return spyObjects;
	}

	/**
	 * @return
	 */
	public Collection getSpyObjectsGC() {
		return spyObjectsGC;
	}

	protected static void printOpenedMethods(
		java.io.PrintStream out,
		JBPThread spyThread) {
		System.out.println("Opened Stack on Thread " + spyThread.getName());
		java.util.Iterator iter = spyThread.getSpyCalls().iterator();
		while (iter.hasNext()) {
			JBPCall spyCall = (JBPCall) iter.next();
			System.out.println("method = " + spyCall.getSpyMethod());
		}
	}

        /**
         * Creates a fileCollection
         */
        private FileCollection createFileCollection(String strName, Class managedClass) throws IOException, FileNotFoundException {
		File fileContent = File.createTempFile(strName,"content");
		fileContent.deleteOnExit();

		File fileIndex = File.createTempFile(strName,"index");
		fileIndex.deleteOnExit();

		RandomAccessFile randomContent = new RandomAccessFile(fileContent,"rw");
		RandomAccessFile managedIndex = new RandomAccessFile(fileIndex,"rw");
		return new FileCollection(randomContent,managedIndex,managedClass,false,0);
	}

	/** Consolida a arvore de chamadas entre as Threads */
	public JBPMethodCount consolidaCallings(JBPMethod method) {
		Iterator methodThreads = method.getThreadList().iterator();
		JBPMethodCount countConsolidado = new JBPMethodCount(method);
		while (methodThreads.hasNext()) {
			JBPMethodThread threadXref = (JBPMethodThread) methodThreads.next();
			JBPThread thread = threadXref.getSpyThread();
			JBPMethod methodOnThread =
				(JBPMethod) thread.getSpyMethods().get(method);

			Iterator counts = methodOnThread.getSpyMethodCount().iterator();
			while (counts.hasNext()) {
				JBPMethodCount spyMethodCountTmp =
					(JBPMethodCount) counts.next();
				if (inDebug)
					spyMethodCountTmp.printDebugCalleds(0);
				copyCount(countConsolidado, spyMethodCountTmp);
			}
		}
		return countConsolidado;
	}

	/** Consolidate all roots on Threads */
	public HashMap consolidateRoots() {
		HashMap roots = new HashMap();
		Iterator threadsIterator = this.spyThreads.values().iterator();
		while (threadsIterator.hasNext()) {
			JBPThread spyThread = (JBPThread) threadsIterator.next();
			copyRootsOnThread(spyThread, roots);
		}
		return roots;
	}

	/** Efetua a copia da arvore */
	private void copyCount(
		JBPMethodCount countConsolidado,
		JBPMethodCount spyMethodCountToCopy) {
		countConsolidado.sum(spyMethodCountToCopy);

		// copiando informacoes sobre objetos criados
		Iterator iterMemoryCount =
			spyMethodCountToCopy.getCountMemory().values().iterator();
		while (iterMemoryCount.hasNext()) {
			JBPMemoryCount memCount = (JBPMemoryCount) iterMemoryCount.next();
			countConsolidado
				.getCountForClass((JBPClass) memCount.getReferencedObject())
				.sum(memCount);
		}

		// Tratamento calleds
		Iterator iterCalled =
			spyMethodCountToCopy.getCountCalleds().values().iterator();
		while (iterCalled.hasNext()) {
			JBPMethodCount methodCountChildToCopy =
				(JBPMethodCount) iterCalled.next();
			JBPMethodCount methodCountChildConsolidado =
				(JBPMethodCount) countConsolidado.getCountCalleds().get(
					methodCountChildToCopy.getSpyMethod());

			if (methodCountChildConsolidado == null) {
				JBPMethod methodChildOnProcess =
					(JBPMethod) this.getSpyMethods().get(
						methodCountChildToCopy.getSpyMethod());
				methodCountChildConsolidado =
					new JBPMethodCount(methodChildOnProcess);
				countConsolidado.getCountCalleds().put(
					methodChildOnProcess,
					methodCountChildConsolidado);
			}

			copyCount(methodCountChildConsolidado, methodCountChildToCopy);

		}

		/*Iterator iterCaller = spyMethodCountToCopy.getCountCallers().values().iterator();
		while (iterCaller.hasNext()) {
		   JBPMethodCount methodCountChildToCopy = (JBPMethodCount)iterCalled.next();
		   JBPMethodCount methodCountChildConsolidado= (JBPMethodCount)countConsolidado.getCountCalleds().get(methodCountChildToCopy.getSpyMethod());

		   if (methodCountChildConsolidado==null) {
		      JBPMethod methodChildOnProcess = this.getSpyMethods().get(methodCountChildConsolidado.getSpyMethod());
		      methodCountChildConsolidado = new JBPMethodCount(methodChildOnProcess);
		      countConsolidado.getCountCalleds().put(methodChildOnProcess,methodCountChildConsolidado);
		   }

		   copyCount(methodCountChildConsolidado,methodCountChildToCopy);

		}*/

	}

	private void copyRootsOnThread(JBPThread thread, HashMap hashToCopy) {
		Iterator rootsOnThreads =
			thread.getSpyMethodCountRoots().values().iterator();
		while (rootsOnThreads.hasNext()) {
			JBPMethodCount countRoot = (JBPMethodCount) rootsOnThreads.next();
			JBPMethodCount countRootTarget =
				(JBPMethodCount) hashToCopy.get(countRoot.getSpyMethod());
			if (countRootTarget == null) {
				countRootTarget = new JBPMethodCount(countRoot.getSpyMethod());
				hashToCopy.put(countRoot.getSpyMethod(), countRootTarget);
			}
			copyCount(countRootTarget, countRoot);
		}

	}

	public boolean equals(Object obj) {
		if (date != null && (((JBPProcess) obj).date) != null) {
			return date.equals(((JBPProcess) obj).date);
		} else {
			return date == (((JBPProcess) obj).date);
		}

	}


	public JBPClass findClassByName(String name) {
	    Iterator iterClasses = this.spyClasses.values().iterator();
	    while (iterClasses.hasNext()) {
	        JBPClass jbpClass = (JBPClass)iterClasses.next();
	        if (jbpClass.getName().equals(name)) {
	            return jbpClass;
	        }
	    }
	    return null;
	}

	public JBPClass findClass(int classId, JBPClass tmpClass) {
		tmpClass.setClassID(classId);
                JBPClass spyClass = (JBPClass)this.spyClasses.get(tmpClass);
		return spyClass;
	}

	/**
         * Look at spyObjectsIndex summarizing where alived instances were created.
	 * @return an ArrayList of {@link JBPMemoryCount}
         */
	public Collection findMemoryLeaks(JBPClass spyClass) {
                HashMap listOfRoots = (HashMap)spyObjectTreeByClass.get(spyClass);
                if (listOfRoots==null) return new ArrayList();
                else return listOfRoots.values();
	}

	/** Finds a method based on its ID, using a temporary instance of JBPMethod which you should keep it cached,
	 *  as we want to save memory and the needs of GC which are large */
	public JBPMethod findMethod(JBPMethod tmpKey, int methodId) {
		tmpKey.setMethodID(methodId);
		return (JBPMethod)spyMethods.get(tmpKey);
	}

	public JBPMethod findMethodByFullName(String fullname) {
		Iterator iterMethods = this.getSpyMethods().values().iterator();
		while (iterMethods.hasNext()) {
			JBPMethod method = (JBPMethod) iterMethods.next();
			if (method.getFullName().equals(fullname)) {
				return method;
			}
		}
		return null;
	}


	public JBPThread findThread(int threadId, JBPThread tmpkeyThread) {
		tmpkeyThread.setThreadId(threadId);
		JBPThread thread =  (JBPThread)spyThreads.get(tmpkeyThread);
		return thread;
	}

	public JBPMemoryState getCurrentMemoryState() {
		return getMemoryState(memoryStates.size()-1);
	}

	public JBPMemoryState getMemoryState(int index) {
		return (JBPMemoryState) memoryStates.get(index);
	}

	public Date getDate() {
		return date;
	}
	public java.util.Date getFinishDate() {
		return finishDate;
	}

	public ArrayList getMemoryStates() {
		return memoryStates;
	}

	public int getNumberOfGarbageCollects() {
		return numberOfGarbageCollects;
	}

	public String getPid() {
		return pid;
	}

	public JBPThread getRootThread() {
		return rootThread;
	}

	public HashMap getSpyClasses() {
		return spyClasses;
	}

	public HashMap getSpyMethods() {
		return spyMethods;
	}

	/**
	 * @return
	 */
	public HashMap getSpyObjectTreeByClass() {
		return spyObjectTreeByClass;
	}

	public HashMap getSpyThreads() {
		return spyThreads;
	}
	public long getTotalClocks() {
		return totalClocks;
	}
	public long getTotalElapsedClocks() {
		return totalElapsedClocks;
	}

	public long getTotalTimeOfGCMilliseconds() {
		return totalTimeOfGCMilliseconds;
	}

	public int hashCode() {
		return date.hashCode();
	}

        /** Creates the indexes HashMap spyObjectsIndex */
	public void indexObjects() {
		throw new RuntimeException ("Do not use this method anymore");
	}

	public void printDebug(java.io.PrintStream out) {
		printDebug(out, "thread = ", spyThreads);
		printDebug(out, "class = ", spyClasses);
		printDebug(out, "method = ", spyMethods);
		System.out.println("RootThread");
		if (rootThread != null) {
			rootThread.printDebug(out);
			printOpenedMethods(out);
		}

	}

	protected void printDebug(
		java.io.PrintStream out,
		String message,
		HashMap hashToDebug) {
		java.util.Iterator threadsIterator = hashToDebug.values().iterator();
		while (threadsIterator.hasNext()) {
			Object obj = threadsIterator.next();
			System.out.println(message + obj);
			if (obj instanceof PrintDebug) {
				((PrintDebug) obj).printDebug(out);
			}
		}

	}

	protected void printOpenedMethods(java.io.PrintStream out) {
		System.out.println();
		System.out.println(
			"******************** Opened Threads ********************");
		System.out.println();

		if (rootThread != null && !rootThread.isClosed()) {
			printOpenedMethods(out, rootThread);
		}

		Iterator threadsIterator = spyThreads.values().iterator();
		while (threadsIterator.hasNext()) {
			JBPThread spyThread = (JBPThread) threadsIterator.next();
			if (!spyThread.isClosed()) {
				printOpenedMethods(out, spyThread);
			}
		}

	}

	public void setDate(Date date) {
		this.date = date;
	}
	public void setFinishDate(java.util.Date finishDate) {
		this.finishDate = finishDate;
	}

	public void setNumberOfGarbageCollects(int numberOfGarbageCollects) {
		this.numberOfGarbageCollects = numberOfGarbageCollects;
	}
	public void setPid(String pid) {
		this.pid = pid;
	}
	public void setTotalClocks(long totalClocks) {
		this.totalClocks = totalClocks;
	}
	public void setTotalElapsedClocks(long totalElapsedClocks) {
		this.totalElapsedClocks = totalElapsedClocks;
	}

	public void setTotalTimeOfGCMilliseconds(long totalTimeOfGCMilliseconds) {
		this.totalTimeOfGCMilliseconds = totalTimeOfGCMilliseconds;
	}

	/**
         * Look at spyObjectsIndex summarizing where alived instances were
         * created.  This method is not part of the interface, and is intended
         * to the process itself.  This is the why we have weird parameters
         * here, regarding more about the counting process than the interface
         * itself.  We might have to revisit this method in the future, trying
         * to put it in a different place.
         *
	 * @param tmpKey A temporary variable to be used when find spyMethods inside HashMaps.
	 * @return an ArrayList of {@link JBPMemoryCount}
         */
	public void summarizeObject(JBPObject spyObject,JBPMethod tmpKey, JBPThread tmpKeyThread, JBPClass tmpClass, boolean isRelease) {
        JBPClass spyClassObject = findClass(spyObject.getClassId(), tmpClass);

        // first step - Add callings into three
		HashMap hashRoots = (HashMap)spyObjectTreeByClass.get(spyClassObject);
		if (hashRoots==null) {
			hashRoots = new HashMap();
			spyObjectTreeByClass.put(spyClassObject,hashRoots);
		}
		int path[] = spyObject.getGraphPathCreation();
		if (path==null || path.length==0) return;

		JBPMethod rootMethod = findMethod(tmpKey,path[0]);

		JBPMethodCount rootCount = (JBPMethodCount)hashRoots.get(rootMethod);
		if (rootCount==null) {
			rootCount = new JBPMethodCount(rootMethod);
			hashRoots.put(rootMethod,rootCount);
		}
		if (isRelease) {
			rootCount.addRelease();
		} else {
			rootCount.add();
		}

		JBPMethodCount currentCount = rootCount;

		for (int i=1;i<path.length;i++) {
			JBPMethod method = findMethod(tmpKey, path[i]);
			currentCount = currentCount.getCalledCount(method,true);
			if (isRelease) {
				currentCount.addRelease();
			} else {
				currentCount.add();
			}
		}

		// Thread's view
		if (isRelease) {
			JBPThread spyThread = findThread(spyObject.getThreadIdCreated(),tmpKeyThread);

			JBPMethodCount methodCount = spyThread.findMethodPath(path,this,tmpKey);
			JBPMemoryCount memoryCount = methodCount.getCountForClass(spyClassObject);
			memoryCount.addRelease(spyObject.getBytesAllocated());

		}
	}

}
