/*
* JBoss, Home of Professional Open Source
* Copyright 2005, JBoss Inc., and individual contributors as indicated
* by the @authors tag. See the copyright.txt in the distribution for a
* full listing of individual contributors.
*
* This is free software; you can redistribute it and/or modify it
* under the terms of the GNU Lesser General Public License as
* published by the Free Software Foundation; either version 2.1 of
* the License, or (at your option) any later version.
*
* This software is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
* Lesser General Public License for more details.
*
* You should have received a copy of the GNU Lesser General Public
* License along with this software; if not, write to the Free
* Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
* 02110-1301 USA, or see the FSF site: http://www.fsf.org.
*/
package org.jboss.profiler;

import org.jboss.aop.pointcut.ast.ParseException;
import org.jboss.profiler.aop.monitoring.control.MonitoringController;
import org.jboss.profiler.aop.monitoring.model.MonitoringMeasure;
import org.jboss.profiler.aop.monitoring.model.TimeValue;

/**
 * @author Clebert Suconic
 */
public class AOPStatisticsMonitoringService implements AOPStatisticsMonitoringServiceMBean {


    String currentMeasure;
    String pointcut;
    boolean started;

    MonitoringMeasure cacheMeasure;


    /* (non-Javadoc)
     * @see org.jboss.profiler.AOPStatisticsMonitoringServiceMBean#getAvailableMeasures()
     */
    public String[] getAvailableMeasures() {
        // Using arrays just as a precaution against multi-thread
        Object[] measures = MonitoringController.getMeasuresSet().toArray();

        String measuresString[] = new String[measures.length];
        for (int i=0;i<measures.length;i++) {
            measuresString[i] = ((MonitoringMeasure)measures[i]).getName();
        }
        return measuresString;
    }

    private MonitoringMeasure getMeasure() {
        if (cacheMeasure!=null) {
            return cacheMeasure;
        } else {
            return MonitoringController.findMeasureByName(currentMeasure);
        }
    }

    private TimeValue getMeasureCurrentGrade() {
        MonitoringMeasure measure = getMeasure();
        if (measure!=null) {
            return MonitoringController.findCurrentTimeValue(measure);
        } else {
            return null;
        }

    }

    /* (non-Javadoc)
     * @see org.jboss.profiler.AOPStatisticsMonitoringServiceMBean#getAvgTime()
     */
    public long getAvgTime() {
        MonitoringMeasure measure = getMeasure();
        if (measure==null) {
            return 0l;
        } else {
            return measure.getRootValue().getAvgTime();
        }
    }

    /* (non-Javadoc)
     * @see org.jboss.profiler.AOPStatisticsMonitoringServiceMBean#getCurrentGradeAvgTime()
     */
    public long getCurrentGradeAvgTime() {
        TimeValue value = getMeasureCurrentGrade();
        if (value==null) {
            return 0;
        } else {
            return value.getAvgTime();
        }
    }
    /* (non-Javadoc)
     * @see org.jboss.profiler.AOPStatisticsMonitoringServiceMBean#getCurrentGradeMaxThreads()
     */
    public long getCurrentGradeMaxThreads() {
        TimeValue value = getMeasureCurrentGrade();
        if (value==null) {
            return 0l;
        } else {
            return value.getMaxStack();
        }
    }
    /* (non-Javadoc)
     * @see org.jboss.profiler.AOPStatisticsMonitoringServiceMBean#getCurrentGradeRequests()
     */
    public long getCurrentGradeCallings() {
        TimeValue value = getMeasureCurrentGrade();
        if (value==null) {
            return 0l;
        } else {
            return value.getCallings();
        }
    }
    /* (non-Javadoc)
     * @see org.jboss.profiler.AOPStatisticsMonitoringServiceMBean#getRequests()
     */
    public long getCallings() {
        MonitoringMeasure measure = getMeasure();
        if (measure==null) {
            return 0l;
        } else {
            return measure.getRootValue().getCallings();
        }
    }
    /* (non-Javadoc)
     * @see org.jboss.profiler.AOPStatisticsMonitoringServiceMBean#maxThreads()
     */
    public long getMaxThreads() {
        MonitoringMeasure measure = getMeasure();
        if (measure==null) {
            return 0;
        } else {
            return measure.getRootValue().getMaxStack();
        }
    }
    /**
     * @return Returns the currentMeasure.
     */
    public String getCurrentMeasure() {
        return currentMeasure;
    }
    /**
     * @param currentMeasure The currentMeasure to set.
     */
    public void setCurrentMeasure(String currentMeasure) {
        cacheMeasure = null;
        this.currentMeasure = currentMeasure;
    }
    /**
     * @return Returns the started.
     */
    public boolean isStarted() {
        return started;
    }

    /* (non-Javadoc)
     * @see org.jboss.profiler.AOPStatisticsMonitoringServiceMBean#getPointcut()
     */
    public String getPointcut() {
        return this.pointcut;
    }

    /* (non-Javadoc)
     * @see org.jboss.profiler.AOPStatisticsMonitoringServiceMBean#setPointcut(java.lang.String)
     */
    public void setPointcut(String ppointcut) throws ParseException {
        if (started && ppointcut!=null && (this.pointcut==null || !ppointcut.equals(this.pointcut))) {
            invokePointcut(ppointcut);
        }
        this.pointcut=ppointcut;

    }
    /* (non-Javadoc)
     * @see org.jboss.profiler.AOPStatisticsMonitoringServiceMBean#start()
     */
    public void start() throws ParseException {
        this.started=true;
        this.invokePointcut(pointcut);
    }

    /** Add the pointcut for this mbean
     * @throws ParseException*/
    private void invokePointcut(String ppointcut) throws ParseException {
        if (ppointcut!=null) {
            System.out.println("Adding pointcut = " + ppointcut);
            new AOPMonitoringService().addPointcut(ppointcut);
        }
    }
}
