/*
* JBoss, Home of Professional Open Source
* Copyright 2005, JBoss Inc., and individual contributors as indicated
* by the @authors tag. See the copyright.txt in the distribution for a
* full listing of individual contributors.
*
* This is free software; you can redistribute it and/or modify it
* under the terms of the GNU Lesser General Public License as
* published by the Free Software Foundation; either version 2.1 of
* the License, or (at your option) any later version.
*
* This software is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
* Lesser General Public License for more details.
*
* You should have received a copy of the GNU Lesser General Public
* License along with this software; if not, write to the Free
* Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
* 02110-1301 USA, or see the FSF site: http://www.fsf.org.
*/

package org.jboss.profiler.filecollection;

import java.io.DataInput;
import java.io.DataOutput;
import java.io.File;
import java.io.IOException;
import java.io.RandomAccessFile;
import java.util.Collection;
import java.util.Comparator;
import java.util.Iterator;


import junit.framework.TestCase;

class TestContainer implements FileCollectionInterface
{
   int number;

    public void readObject(DataInput ch) throws IOException
    {
        number = ch.readInt();
    }

    public void writeObject(DataOutput ch) throws IOException
    {
        ch.writeInt(number);
    }

}

class TestComparator implements Comparator
{

    public int compare(Object arg0, Object arg1)
    {
        TestContainer cont0=(TestContainer)arg0;
        TestContainer cont1=(TestContainer)arg1;
        if (cont0.number==cont1.number)
        {
            return 0;
        } else
        if (cont0.number>cont1.number)
        {
            return 1;
        } else
        {
            return -1;
        }
        //return ((TestContainer)arg0).number - ((TestContainer)arg1).number;
    }

}

/**
 * @author csuconic
 */
public class FileCollectionTest extends TestCase
{
    public void testFixed() throws Exception
    {
        File fileData = File.createTempFile("tst",".data");
        File fileIndex = File.createTempFile("tst",".index");
        RandomAccessFile access = new RandomAccessFile(fileData,"rw");
        RandomAccessFile index = new  RandomAccessFile(fileIndex,"rw");
        fileData.deleteOnExit();
        fileIndex.deleteOnExit();
        FileCollection fileColl = new FileCollection(access,index, TestContainer.class,false,100);


        for (int i=255;i>0;i--)
        {
            TestContainer container = new TestContainer();
            container.number=i;
            fileColl.add(container);
        }

        Iterator iter = fileColl.iterator();

        int i=255;
        while (iter.hasNext())
        {
            TestContainer container = (TestContainer)iter.next();
            assertEquals("Invalid sequence",i,container.number);
            i--;
        }


        for (i=0;i<255;i++)
        {
            TestContainer container = new TestContainer();
            container.number=i;
            fileColl.set(i,container);
        }

        iter = fileColl.iterator();

        i=0;
        while (iter.hasNext())
        {
            TestContainer container = (TestContainer)iter.next();
            assertEquals("Invalid sequence",i,container.number);
            i++;
        }

        for (i=0;i<255;i++)
        {
            TestContainer container = (TestContainer)fileColl.get(i);
            assertEquals("Invalid sequence",i,container.number);
        }

        access.close();
        index.close();
    }

    public void testBulkAdd() throws Exception
    {
        File fileData = File.createTempFile("tst",".bulkdata");
        File fileIndex = File.createTempFile("tst",".bulkindex");
        RandomAccessFile access = new RandomAccessFile(fileData,"rw");
        RandomAccessFile index = new  RandomAccessFile(fileIndex,"rw");
        //fileData.deleteOnExit();
        //fileIndex.deleteOnExit();
        FileCollection fileColl = new FileCollection(access,index, TestContainer.class,false,100);
        BulkAdd bulkAdd = fileColl.createBulkAdd();

        for (int i=5000;i>0;i--)
        {
            TestContainer container = new TestContainer();
            container.number=i;
            bulkAdd.add(container);
        }
        bulkAdd.close();

        Iterator iter = fileColl.iterator();

        int i=5000;
        while (iter.hasNext())
        {
            TestContainer container = (TestContainer)iter.next();
            assertEquals("Invalid sequence",i,container.number);
            i--;
        }


        for (i=0;i<255;i++)
        {
            TestContainer container = new TestContainer();
            container.number=i;
            fileColl.set(i,container);
        }

        iter = fileColl.iterator();

        i=0;
        while (iter.hasNext())
        {
            TestContainer container = (TestContainer)iter.next();
            assertEquals("Invalid sequence",i,container.number);
            i++;
            if (i>=255)
            {
                break;
            }
        }

        for (i=0;i<255;i++)
        {
            TestContainer container = (TestContainer)fileColl.get(i);
            assertEquals("Invalid sequence",i,container.number);
        }

        access.close();
        index.close();


    }


    public void testQuicksort() throws Exception
    {
        File fileData = File.createTempFile("tst",".bigdata");
        File fileIndex = File.createTempFile("tst",".bigindex");
        RandomAccessFile access = new RandomAccessFile(fileData,"rw");
        //RandomAccessFile index = new  RandomAccessFile(fileIndex,"rw");
        //fileData.deleteOnExit();
        //fileIndex.deleteOnExit();
        FileCollection fileColl = new FileCollection(access,null, TestContainer.class,false,100);
        for (int i=5000;i>0;i--)
        {
            TestContainer container = new TestContainer();
            container.number=i;
            fileColl.add(container);
        }

        Iterator iter = fileColl.iterator();

        QuickSorter sorter = new QuickSorter(new TestComparator());

        //printData(fileColl,"Before sort = ");

        sorter.sort(fileColl);
        assertTrue("is Sorted", sorter.isSorted(fileColl));


        //printData(fileColl,"After sort data = ");

        iter = fileColl.iterator();

        int i=0;
        while (iter.hasNext())
        {
            TestContainer container = (TestContainer)iter.next();
            assertEquals("Invalid sequence",i+1,container.number);
            i++;
        }

        access.close();
        //index.close();
    }

    private void printData(FileCollection fileColl, String text)
    {
        Iterator iter;
        iter = fileColl.iterator();
        while (iter.hasNext())
        {
            TestContainer container = (TestContainer)iter.next();
            System.out.println(text + container.number);
        }
    }

    public void testFixedLookingForFailure() throws Exception
    {
        File fileData = File.createTempFile("tst",".data");
        File fileIndex = File.createTempFile("tst",".index");
        RandomAccessFile access = new RandomAccessFile(fileData,"rw");
        RandomAccessFile index = new  RandomAccessFile(fileIndex,"rw");
        fileData.deleteOnExit();
        fileIndex.deleteOnExit();
        FileCollection fileColl = new FileCollection(access,index, TestContainer.class,false,3);

        try {
            TestContainer data = new TestContainer();
            data.number=10;
            fileColl.add(data);
            fail("Supposed to throw an exception");
        }
        catch(Exception ignored)
        {

        }

        access.close();
        index.close();


    }

    public void testRegular() throws Exception
    {
        File fileData = File.createTempFile("tst",".data");
        File fileIndex = File.createTempFile("tst",".index");
        RandomAccessFile access = new RandomAccessFile(fileData,"rw");
        RandomAccessFile index = new  RandomAccessFile(fileIndex,"rw");
        fileData.deleteOnExit();
        fileIndex.deleteOnExit();
        FileCollection fileColl = new FileCollection(access,index, TestContainer.class,false,0);
        for (int i=255;i>0;i--)
        {
            TestContainer container = new TestContainer();
            container.number=i;
            fileColl.add(container);
        }

        Iterator iter = fileColl.iterator();

        int i=255;
        while (iter.hasNext())
        {
            TestContainer container = (TestContainer)iter.next();
            assertEquals("Invalid sequence",i,container.number);
            i--;
        }

        access.close();
        index.close();
    }

    public void testRegularReusing() throws Exception
    {
        File fileData = File.createTempFile("tst",".data");
        File fileIndex = File.createTempFile("tst",".index");
        RandomAccessFile access = new RandomAccessFile(fileData,"rw");
        RandomAccessFile index = new  RandomAccessFile(fileIndex,"rw");
        //fileData.deleteOnExit();
        fileIndex.deleteOnExit();
        FileCollection fileColl = new FileCollection(access,index, TestContainer.class,true,0);
        for (int i=255;i>0;i--)
        {
            TestContainer container = new TestContainer();
            container.number=i;
            fileColl.add(container);
        }

        Iterator iter = fileColl.iterator();

        int i=255;
        while (iter.hasNext())
        {
            TestContainer container = (TestContainer)iter.next();
            assertEquals("Invalid sequence",i,container.number);
            i--;
        }

        access.close();
        index.close();
    }


    public void testFinding() throws Exception
    {
        File fileData = File.createTempFile("tst",".bigdata");
        File fileIndex = File.createTempFile("tst",".bigindex");
        RandomAccessFile access = new RandomAccessFile(fileData,"rw");
        //RandomAccessFile index = new  RandomAccessFile(fileIndex,"rw");
        //fileData.deleteOnExit();
        //fileIndex.deleteOnExit();
        FileCollection fileColl = new FileCollection(access,null, TestContainer.class,false,100);
        for (int i=1000;i>0;i--)
        {
            TestContainer container = new TestContainer();
            container.number=i;
            for (int j=0;j<10;j++)
            {
                fileColl.add(container);
            }
        }

        QuickSorter sorter = new QuickSorter(new TestComparator());

        //printData(fileColl,"Before sort = ");

        sorter.sort(fileColl);
        assertTrue("is Sorted", sorter.isSorted(fileColl));

        TestContainer key = new TestContainer();
        key.number=1000;
        Collection result = sorter.binarySearchCollection(fileColl,key);

        assertEquals(10,result.size());

        Iterator iter = result.iterator();
        while (iter.hasNext())
        {
            TestContainer container = (TestContainer)iter.next();
            assertEquals(1000,container.number);
        }
    }

}

