/*
* JBoss, Home of Professional Open Source
* Copyright 2005, JBoss Inc., and individual contributors as indicated
* by the @authors tag. See the copyright.txt in the distribution for a
* full listing of individual contributors.
*
* This is free software; you can redistribute it and/or modify it
* under the terms of the GNU Lesser General Public License as
* published by the Free Software Foundation; either version 2.1 of
* the License, or (at your option) any later version.
*
* This software is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
* Lesser General Public License for more details.
*
* You should have received a copy of the GNU Lesser General Public
* License along with this software; if not, write to the Free
* Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
* 02110-1301 USA, or see the FSF site: http://www.fsf.org.
*/
#include "liteProfiler.h"

#ifdef CYGWIN
#define WINDOWS
#endif

char * host = NULL;
char * serverName = NULL;
void startProducer();

// global jvmpi interface pointer
static JVMPI_Interface *jvmpi_interface = NULL;
static JNIEnv * jni_interface = NULL;

#define CALL(f) (jvmpi_interface->f)

static std::vector<ThreadData *> validInstances;



void addThreadInstance(ThreadData * data) {
	validInstances.insert(validInstances.end(),data);
}

void verifyHandler(JNIEnv *env_id, ThreadData** data) {
	*data = (ThreadData *)(CALL(GetThreadLocalStorage)(env_id));
	if (((void *)(*data))==NULL) {
		*data = new ThreadData(env_id);
		CALL(SetThreadLocalStorage)(env_id, *data);
		addThreadInstance(*data);
	}
}


void objectAlloc(JNIEnv * env_id, jint size) {
	ThreadData *data = NULL;
	verifyHandler(env_id,&data);
	data->allocatedMemory += size;
	data->objectsAllocated++;
}

void objectRelease(JNIEnv * env_id) {
	ThreadData *data = NULL;
	verifyHandler(env_id,&data);
	data->objectsReleased++;
}

void updateCPU(JNIEnv *env_id) {
	ThreadData *data;
	verifyHandler(env_id,&data);
	data->usedCPU = CALL(GetCurrentThreadCpuTime)();
}

void threadEnd(JNIEnv *env_id) {
	ThreadData *data;
	verifyHandler(env_id,&data);
	data->usedCPU = CALL(GetCurrentThreadCpuTime)();
	data->running=0;
}

// function for handling event notification
void notifyEvent(JVMPI_Event *event) {


  switch(event->event_type) {
  case JVMPI_EVENT_METHOD_EXIT:
    // we will try to get rid of this event. And use this information in a different way.
    updateCPU(event->env_id);
    break;

  case JVMPI_EVENT_OBJECT_ALLOC:
         objectAlloc(event->env_id,
		     event->u.obj_alloc.size);
    break;

  case JVMPI_EVENT_OBJECT_FREE:
        objectRelease(event->env_id);
    break;

  case JVMPI_EVENT_THREAD_END:
     threadEnd(event->env_id);
     break;

  case JVMPI_EVENT_JVM_INIT_DONE:
        initEvents();
        startProducer();

        break;
  case JVMPI_EVENT_JVM_SHUT_DOWN:
       CALL(RunGC)();
       //debugThreads();
	   break;

  default:
     fprintf (stderr,"ERROR - DEFAULTWITH ON EVENT -> %d\n",event->event_type);
  }

}


void debugThreads() {
   std::vector<ThreadData *>::iterator p;

   for (p=validInstances.begin();p!=validInstances.end();p++) {
    /*if ((**p).running) {
       updateCPU((**p).env_id);
    }*/
#ifdef WINDOWS
	fprintf(stderr,"CPU=%I64d ",(**p).usedCPU);
#else
	fprintf(stderr,"CPU=%lld ",(**p).usedCPU);
#endif

      fprintf(stderr,"Bytes=%ld ",(**p).allocatedMemory);
      fprintf(stderr,"objectsAllocated=%ld ",(**p).objectsAllocated);
      fprintf(stderr,"objectsReleased=%ld\n",(**p).objectsReleased);
   }
}


// main
// profiler agent entry point
extern "C" {
  JNIEXPORT jint JNICALL JVM_OnLoad(JavaVM *jvm, char *options, void *reserved) {
    // get jvmpi interface pointer
    fprintf (stdout,"Process %ld\n",getpid());

    int n=0;
    char ** parameterStrings = divideStringsByCharacter(options,',',&n);

    if (n!=2) {
       fprintf (stderr,"Use:-XrunjbossLiteProfiler:host,name");
       exit(-1);
    }

    host = parameterStrings[0];
    serverName=parameterStrings[1];


    if ((jvm->GetEnv((void **)&jvmpi_interface, JVMPI_VERSION_1)) < 0) {
      fprintf(stderr, "jbossProfiler> error in obtaining jvmpi interface pointer\n");
      return JNI_ERR;
    }

    if ((jvm->GetEnv((void **)&jni_interface, JNI_VERSION_1_2)) < 0) {
      fprintf(stderr, "jbossProfiler> error in obtaining JNI interface pointer\n");
      return JNI_ERR;
    }

    // initialize jvmpi interface
    jvmpi_interface->NotifyEvent = notifyEvent;

    jvmpi_interface->EnableEvent(JVMPI_EVENT_JVM_INIT_DONE, NULL);
    jvmpi_interface->EnableEvent(JVMPI_EVENT_JVM_SHUT_DOWN, NULL);


    return JNI_OK;
  }
}

void initEvents() {
    jvmpi_interface->EnableEvent(JVMPI_EVENT_JVM_INIT_DONE, NULL);
    jvmpi_interface->EnableEvent(JVMPI_EVENT_THREAD_END, NULL);
}

/*
 * Class:     org_jboss_profiler_liteversion_jni_LiteVersionController
 * Method:    runGC
 * Signature: ()V
 */
JNIEXPORT void JNICALL Java_org_jboss_profiler_liteversion_jni_LiteVersionController_runGC
  (JNIEnv *, jclass) {
  if (jvmpi_interface!=NULL) {
   CALL(RunGC)();
  }
}

/*
 * Class:     org_jboss_profiler_liteversion_jni_LiteVersionController
 * Method:    getCPU
 * Signature: ()J
 */
JNIEXPORT jlong JNICALL Java_org_jboss_profiler_liteversion_jni_LiteVersionController_getCPU
  (JNIEnv *, jclass) {
   std::vector<ThreadData *>::iterator p;

   jlong usedCPU = 0;
   for (p=validInstances.begin();p!=validInstances.end();p++) {
      ThreadData * data = *p;
      usedCPU += (data->usedCPU - data->old_usedCPU);
   }

   return usedCPU;
}

/*
 * Class:     org_jboss_profiler_liteversion_jni_LiteVersionController
 * Method:    getBytes
 * Signature: ()J
 */
JNIEXPORT jlong JNICALL Java_org_jboss_profiler_liteversion_jni_LiteVersionController_getBytes
  (JNIEnv *, jclass) {
   std::vector<ThreadData *>::iterator p;

   jlong allocatedMemory = 0;
   for (p=validInstances.begin();p!=validInstances.end();p++) {
      ThreadData * data = *p;
      allocatedMemory += (data->allocatedMemory - data->old_allocatedMemory);
   }

   return allocatedMemory;
}

/*
 * Class:     org_jboss_profiler_liteversion_jni_LiteVersionController
 * Method:    getCreatedObjects
 * Signature: ()I
 */
JNIEXPORT jint JNICALL Java_org_jboss_profiler_liteversion_jni_LiteVersionController_getCreatedObjects
  (JNIEnv *, jclass) {
   std::vector<ThreadData *>::iterator p;

   jint objectsAllocated = 0;
   for (p=validInstances.begin();p!=validInstances.end();p++) {
      ThreadData * data = *p;
      objectsAllocated += (data->objectsAllocated - data->old_objectsAllocated);
   }

   return objectsAllocated;
}

/*
 * Class:     org_jboss_profiler_liteversion_jni_LiteVersionController
 * Method:    getReleasedObjects
 * Signature: ()I
 */
JNIEXPORT jint JNICALL Java_org_jboss_profiler_liteversion_jni_LiteVersionController_getReleasedObjects
  (JNIEnv *, jclass) {
   std::vector<ThreadData *>::iterator p;

   jint objectsReleased = 0;
   for (p=validInstances.begin();p!=validInstances.end();p++) {
      ThreadData * data = *p;
      objectsReleased += ((data->objectsReleased) - (data->old_objectsReleased));
   }

   return objectsReleased;
}

/*
 * Class:     org_jboss_profiler_liteversion_jni_LiteVersionController
 * Method:    getTotalCPU
 * Signature: ()J
 */
JNIEXPORT jlong JNICALL Java_org_jboss_profiler_liteversion_jni_LiteVersionController_getTotalCPU
  (JNIEnv *, jclass) {
   std::vector<ThreadData *>::iterator p;

   jlong usedCPU = 0;
   for (p=validInstances.begin();p!=validInstances.end();p++) {
      ThreadData * data = *p;
      usedCPU += (data->usedCPU);
   }

   return usedCPU;
}

/*
 * Class:     org_jboss_profiler_liteversion_jni_LiteVersionController
 * Method:    getTotalBytes
 * Signature: ()J
 */
JNIEXPORT jlong JNICALL Java_org_jboss_profiler_liteversion_jni_LiteVersionController_getTotalBytes
  (JNIEnv *, jclass) {
   std::vector<ThreadData *>::iterator p;

   jlong allocatedMemory = 0;
   for (p=validInstances.begin();p!=validInstances.end();p++) {
      ThreadData * data = *p;
      allocatedMemory += (data->allocatedMemory);
   }

   return allocatedMemory;
}

/*
 * Class:     org_jboss_profiler_liteversion_jni_LiteVersionController
 * Method:    getTotalCreatedObjects
 * Signature: ()I
 */
JNIEXPORT jint JNICALL Java_org_jboss_profiler_liteversion_jni_LiteVersionController_getTotalCreatedObjects
  (JNIEnv *, jclass) {
   std::vector<ThreadData *>::iterator p;

   jint objectsAllocated = 0;
   for (p=validInstances.begin();p!=validInstances.end();p++) {
      ThreadData * data = *p;
      objectsAllocated += (data->objectsAllocated);
   }

   return objectsAllocated;
}

/*
 * Class:     org_jboss_profiler_liteversion_jni_LiteVersionController
 * Method:    getTotalReleasedObjects
 * Signature: ()I
 */
JNIEXPORT jint JNICALL Java_org_jboss_profiler_liteversion_jni_LiteVersionController_getTotalReleasedObjects
  (JNIEnv *, jclass) {
   std::vector<ThreadData *>::iterator p;

   jint objectsReleased = 0;
   for (p=validInstances.begin();p!=validInstances.end();p++) {
      ThreadData * data = *p;
      objectsReleased += ((data->objectsReleased));
   }

   return objectsReleased;
}


/*
 * Class:     org_jboss_profiler_liteversion_jni_LiteVersionController
 * Method:    saveCurrentInfo
 * Signature: ()V
 */
JNIEXPORT void JNICALL Java_org_jboss_profiler_liteversion_jni_LiteVersionController_saveCurrentInfo
  (JNIEnv *, jclass) {
   std::vector<ThreadData *>::iterator p;

   for (p=validInstances.begin();p!=validInstances.end();p++) {
      ThreadData * data = *p;
	  data->old_allocatedMemory = data->allocatedMemory;
	  data->old_objectsAllocated = data->objectsAllocated;
	  data->old_objectsReleased  = data->objectsReleased;
	  data->old_usedCPU = data->usedCPU;
   }
}


JNIEXPORT jstring JNICALL Java_org_jboss_profiler_liteversion_jni_LiteVersionController_getName
  (JNIEnv *, jclass) {
  static jstring name = NULL;
  if (name==NULL) {
     if (serverName==NULL) {
        if (jni_interface!=NULL) {
	     	name = jni_interface->NewStringUTF("undefined");
	    }
     } else {
	     name = jni_interface->NewStringUTF(serverName);
	 }
  }

  return name;
}


// rotina que ira conectar com a console
void startProducer() {
     fprintf (stderr,"Starting producer\n");
     jclass checkClass=jni_interface->FindClass("org/jboss/profiler/liteversion/consolidator/DataProducerProxy");
     if (checkClass==NULL) {
            fprintf (stderr,"\n***********************\nFatalError: Cannot find org.jboss.profiler.liteversion.consolidator.DataProducerProxy(String)\n");
            fprintf (stderr,"Check you bootclasspath for profilerSynchronizer.jar\n");
	     return;
     }

     jmethodID metodoInit=jni_interface->GetStaticMethodID(checkClass,"startProducer","(Ljava/lang/String;)V");
     if (metodoInit==NULL) {
        fprintf (stderr,"\n***********************\nFatalError: Cannot find br.com.summatech.serverSpy.threadcheck.StartupProfillerCheck.initCheck(String,int)\n");
        fprintf (stderr,"Check you bootclasspath for profilerConsole.jar\n");
        return;
     }

     jstring strParameter = jni_interface->NewStringUTF(host);
     jni_interface->CallStaticVoidMethod(checkClass,metodoInit,strParameter);

}

/*
 * Class:     org_jboss_profiler_liteversion_jni_LiteVersionController
 * Method:    activate
 * Signature: ()V
 */
JNIEXPORT void JNICALL Java_org_jboss_profiler_liteversion_jni_LiteVersionController_activate
  (JNIEnv *, jclass) {
    jvmpi_interface->EnableEvent(JVMPI_EVENT_METHOD_EXIT, NULL);
    jvmpi_interface->EnableEvent(JVMPI_EVENT_OBJECT_ALLOC, NULL);
    jvmpi_interface->EnableEvent(JVMPI_EVENT_OBJECT_FREE, NULL);
}

/*
 * Class:     org_jboss_profiler_liteversion_jni_LiteVersionController
 * Method:    deActivate
 * Signature: ()V
 */
JNIEXPORT void JNICALL Java_org_jboss_profiler_liteversion_jni_LiteVersionController_deActivate
  (JNIEnv *, jclass){
    jvmpi_interface->DisableEvent(JVMPI_EVENT_METHOD_EXIT, NULL);
    jvmpi_interface->DisableEvent(JVMPI_EVENT_OBJECT_ALLOC, NULL);
    jvmpi_interface->DisableEvent(JVMPI_EVENT_OBJECT_FREE, NULL);
}
